# -*- coding: utf-8 -*-

"""NPT (canonical) dynamics in LAMMPS"""

import json

import lammps_step
import logging
import seamm
from seamm_util import units_class, Q_
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import random

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("lammps")


class NPT(lammps_step.NVT):
    methods = {
        "Nose-Hoover": {
            "documentation": "https://lammps.sandia.gov/doc/fix_nh.html#fix-npt-command",  # noqa: E501
            "references": ["Shinoda", "Tuckerman"],
        },
        "Berendsen": {
            "documentation": "https://lammps.sandia.gov/doc/fix_press_berendsen.html",  # noqa: E501
            "references": ["Berendsen"],
        },
    }

    def __init__(self, flowchart=None, title="NPT dynamics", extension=None):
        """Initialize the node"""

        logger.debug("Creating NPT {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        logger.debug("NPT after super init, {}".format(self))

        self.description = "NPT dynamics step in LAMMPS"

        logger.debug("NPT.init() creating NPT_Parameters object")

        self._calculation = "npt"
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.NPT_Parameters()

        logger.debug("NPT.init() completed")

    def description_text(self, P=None):
        """Create the text description of what this step will do.
        The dictionary of control values is passed in as P so that
        the code can test values, etc.
        """

        if not P:
            P = self.parameters.values_to_dict()

        model = self.model

        # What will we do?

        if P["T0"] == P["T1"]:
            text = "{time} of canonical (NPT) dynamics at {T0} "
        else:
            text = (
                "{time} of canonical (NPT) dynamics starting "
                " at {T0}, going to {T1}, "
            )
        text += "using a timestep of {timestep} "
        if model is None:
            text += ". The temperature will be controlled "
        else:
            text += f"using the {model} forcefield. The temperature will be controlled "
        if P["thermostat"] == "Nose-Hoover":
            text += "using a Nose-Hoover thermostat."
            if P["Tchain"] != "3":
                if P["Tloop"] != "1":
                    text += (
                        " The thermostat will use a chain of {Tchain} "
                        "thermostats with {Tloop} subcycles and a "
                    )
                else:
                    text += (
                        " The thermostat will use a chain of {Tchain} "
                        "thermostats and a "
                    )
            elif P["Tloop"] != "1":
                text += " The thermostat will use {Tloop} subcycles and a "
            else:
                text += " The thermostat will use a "
            text += "drag factor of {drag}."
        elif P["thermostat"] == "Berendsen":
            text += "using a Berendsen thermostat with a damping time " "of {Tdamp}"
        elif "csvr" in P["thermostat"]:
            text += (
                "using a canonical sampling thermostat using velocity "
                "rescaling (CSVR) with a damping time of {Tdamp} and "
                "a {random_seed}."
            )
        elif "csld" in P["thermostat"]:
            text += (
                "using a canonical sampling thermostat using Langevin "
                "dynamics (CSLD) with a damping time of {Tdamp} and "
                "a {random_seed}."
            )
        elif P["thermostat"] == "velocity rescaling":
            text += (
                "using velocity rescaling every {frequency} with a "
                "temperature window of {window}."
            )
            if P["fraction"] != 1.0:
                text += (
                    " The velocities will only be scaled a fraction "
                    "({fraction}) of the amount needed to fully correct "
                    "the temperature."
                )
        elif P["thermostat"] == "Langevin":
            text += (
                "using a Langevin thermostat with a damping time "
                "of {Tdamp} and a {random_seed}."
            )
        else:
            text += "using the thermostat given by {thermostat}."

        text += " The pressure will be controlled using a {barostat} barostat."

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def describe(self, indent="", json_dict=None):
        """Write out information about what this node will do
        If json_dict is passed in, add information to that dictionary
        so that it can be written out by the controller as appropriate.
        """

        # Can't call super() because it will print too much
        self.visited = True
        job.job("\n" + self.indent + self.header)
        next_node = self.next()

        # Local copies of variables in a dictionary

        P = self.parameters.values_to_dict()
        text = self.description_text(P)
        job.job(__(text, indent=self.indent + "    ", **P))

        return next_node

    def get_input(self, extras=None):
        """Get the input for an NPT dynamics run in LAMMPS"""

        self.description = []

        _, configuration = self.get_system_configuration()
        if configuration.periodicity == 0:
            raise RuntimeError("Cannot run NPT dynamics on non-periodic system!")

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Fix variables with special cases
        timestep, P["timestep"] = self.timestep(P["timestep"])

        if P["seed"] == "random":
            # Apparently the seed must be no larger than 900,000,000
            P["seed"] = int(random.random() * 899999999)

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description.append(__(self.description_text(PP), **PP, indent=4 * " "))

        # May need to force a triclinic cell
        if P["allow shear"]:
            self.parent.force_triclinic = True

        time = lammps_step.to_lammps_units(P["time"], quantity="time")
        nsteps = round(time / timestep)

        T0 = lammps_step.to_lammps_units(P["T0"], quantity="temperature")
        T1 = lammps_step.to_lammps_units(P["T1"], quantity="temperature")
        Tdamp = lammps_step.to_lammps_units(P["Tdamp"], quantity="time")

        barostat = P["barostat"]
        if barostat == "Berendsen":
            modulus = lammps_step.to_lammps_units(P["modulus"], quantity="pressure")

        # Work out the pressure/stress part of the command
        ptext = self.get_pressure_text(P)

        thermo_properties = (
            "time temp press etotal ke pe ebond "
            "eangle edihed eimp evdwl etail ecoul elong"
        )

        properties = (
            "v_time v_temp v_press v_vol v_density "
            "v_etotal v_ke v_pe v_epair "
            "v_cella v_cellb v_cellc"
        )
        title2 = "tstep t T P V density Etot Eke Epe Epair a b c"
        if P["allow shear"]:
            properties += " v_cellalpha v_cellbeta v_cellgamma"
            title2 += " alpha beta gamma"
        if configuration.periodicity == 3:
            properties += " v_sxx v_syy v_szz v_syz v_sxz v_sxy"
            title2 += " Sxx Syy Szz Syz Sxz Sxy"
        if self.parent.have_dreiding_hbonds:
            thermo_properties += " v_N_hbond v_E_hbond"
            properties += " v_N_hbond v_E_hbond"
            title2 += " N_hbond E_hbond"

        # and build the LAMMPS script
        lines = []
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")
        lines.append("reset_timestep      0")
        lines.append("timestep            {}".format(timestep))
        lines.append("thermo_style        custom {}".format(thermo_properties))
        lines.append("thermo              {}".format(int(nsteps / 100)))

        ncomputes = 0
        ndumps = 0
        nfixes = 0
        if P["thermostat"] == "Nose-Hoover":
            Tchain = P["Tchain"]
            Tloop = P["Tloop"]
            drag = P["drag"]
            if barostat == "Nose-Hoover":
                nfixes += 1
                lines.append(
                    "fix                 {} all npt ".format(nfixes)
                    + "temp {} {} {} ".format(T0, T1, Tdamp)
                    + "tchain {} ".format(Tchain)
                    + "tloop {} ".format(Tloop)
                    + "drag {}".format(drag)
                    + ptext
                )

                for citation in NPT.methods["Nose-Hoover"]["references"]:
                    self.references.cite(
                        raw=self._bibliography[citation],
                        alias=citation,
                        module="lammps_step",
                        level=1,
                        note="Citation for NPT barostat.",
                    )
            else:
                nfixes += 1
                lines.append(
                    "fix                 {} all nvt ".format(nfixes)
                    + "temp {} {} {} ".format(T0, T1, Tdamp)
                    + "tchain {} ".format(Tchain)
                    + "tloop {} ".format(Tloop)
                    + "drag {}".format(drag)
                )
                nfixes += 1
                lines.append(
                    "fix                 {} all ".format(nfixes)
                    + "press/berendsen "
                    + ptext
                    + " modulus {}".format(modulus)
                )

                for citation in NPT.methods["Berendsen"]["references"]:
                    self.references.cite(
                        raw=self._bibliography[citation],
                        alias=citation,
                        module="lammps_step",
                        level=1,
                        note="Citation for NPT barostat.",
                    )
        elif P["barostat"] == "Berendsen":
            nfixes += 1
            lines.append(
                "fix                 {} all ".format(nfixes)
                + "press/berendsen "
                + ptext
                + " modulus {}".format(modulus)
            )

            for citation in NPT.methods["Berendsen"]["references"]:
                self.references.cite(
                    raw=self._bibliography[citation],
                    alias=citation,
                    module="lammps_step",
                    level=1,
                    note="Citation for NPT barostat.",
                )
            if P["thermostat"] == "Berendsen":
                nfixes += 1
                lines.append(
                    "fix                 {} ".format(nfixes)
                    + "all temp/berendsen "
                    + " {} {} {}".format(T0, T1, Tdamp)
                )
                nfixes += 1
                lines.append("fix                 {} ".format(nfixes) + "all nve")
            elif "csvr" in P["thermostat"]:
                seed = P["seed"]
                nfixes += 1
                lines.append(
                    "fix                 {} ".format(nfixes)
                    + "all temp/csvr "
                    + " {} {} {} {}".format(T0, T1, Tdamp, seed)
                )
                nfixes += 1
                lines.append("fix                 {} ".format(nfixes) + "all nve")
            elif "csld" in P["thermostat"]:
                seed = P["seed"]
                nfixes += 1
                lines.append(
                    "fix                 {} ".format(nfixes)
                    + "all temp/csld "
                    + " {} {} {} {}".format(T0, T1, Tdamp, seed)
                )
                nfixes += 1
                lines.append("fix                 {} ".format(nfixes) + "all nve")
            elif P["thermostat"] == "velocity rescaling":
                frequency = P["frequency"]
                nevery = max(1, round(nsteps / (frequency / timestep)))
                window = lammps_step.to_lammps_units(
                    P["window"], quantity="temperature"
                )
                fraction = P["fraction"]
                nfixes += 1
                lines.append(
                    "fix                 {} ".format(nfixes)
                    + "all temp/rescale "
                    + "{} {} {} {} {}".format(nevery, T0, T1, window, fraction)
                )
                nfixes += 1
                lines.append("fix                 {} ".format(nfixes) + "all nve")
            elif P["thermostat"] == "Langevin":
                seed = P["seed"]
                nfixes += 1
                lines.append(
                    "fix                 {} ".format(nfixes)
                    + "all langevin "
                    + "{} {} {} {} ".format(T0, T1, Tdamp, seed)
                )
                nfixes += 1
                lines.append("fix                 {} ".format(nfixes) + "all nve")
            else:
                raise RuntimeError(
                    f"Don't recognize temperature control {P['thermostat']}"
                )

            # Add the citation for the thermostat
            metadata = lammps_step.thermostat_metadata
            for citation in metadata[P["thermostat"]]["references"]:
                self.references.cite(
                    raw=self._bibliography[citation],
                    alias=citation,
                    module="lammps_step",
                    level=1,
                    note="Citation for thermostat.",
                )

        # For the heat flux, if requested, we need extra input
        if P["heat flux"] != "never":
            # Unit conversion factor
            if lammps_step.get_lammps_unit_system() == "metal":
                factor = Q_("eV/Å^2/ps")
            else:
                factor = (
                    Q_("kcal/Å^2/fs/mol") / Q_("kcal/mol") * Q_("kcal/mol").to("kJ")
                )
            factor = factor.m_as("W/m^2")
            lines.append(
                f"""
compute             KE all ke/atom
compute             PE all pe/atom

#          centroid doesn't work with kspace, so split into pair and non-pair parts

compute             S_p all stress/atom NULL pair kspace
compute             S_b all centroid/stress/atom NULL bond angle dihedral improper
compute             flux_p all heat/flux KE PE S_p
compute             flux_b all heat/flux KE PE S_b

#          Conversion from kcal/Å^2/fs/mol to W/m^2")

variable            factor equal {factor}
variable            Jx equal v_factor*(c_flux_p[1]+c_flux_b[1])/vol
variable            Jy equal v_factor*(c_flux_p[2]+c_flux_b[2])/vol
variable            Jz equal v_factor*(c_flux_p[3]+c_flux_b[3])/vol
"""
            )

        # summary output written 10 times during run so we can see progress
        nevery = 10
        nfreq = int(nsteps / 10)
        nrepeat = int(nfreq / nevery)
        nfreq = nevery * nrepeat
        nfixes += 1
        filename = f"@{self._id[-1]}+npt_summary.trj"
        lines.append(
            f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
            f"                       {properties} &\n"
            "                       off 2 &\n"
            f"                       file {filename}"
        )
        # instantaneous output written for averaging
        if P["sampling"] == "none":
            self.description.append(
                __(
                    "The run will be {nsteps:n} steps of dynamics.",
                    nsteps=nsteps,
                    indent=4 * " ",
                )
            )
        else:
            sampling = lammps_step.to_lammps_units(P["sampling"], quantity="time")
            nevery = max(1, round(sampling / timestep))
            nfreq = int(nsteps / nevery)
            nrepeat = 1
            nfreq = nevery * nrepeat
            nfixes += 1
            dt = (nevery * P["timestep"]).to_compact()
            P0tmp = P["Pinitial"].to_compact()
            P1tmp = P["Pfinal"].to(P0tmp.units)
            tmp = {
                "code": "LAMMPS",
                "type": "NPT",
                "dt": dt.magnitude,
                "tunits": str(dt.u),
                "nsteps": nsteps // nevery,
                "Tunits": "K",
                "Punits": str(P0tmp.u),
            }
            if P0tmp == P1tmp:
                tmp["P"] = P0tmp.magnitude
            else:
                tmp["P0"] = P0tmp.magnitude
                tmp["T1"] = P1tmp.magnitude
            if T0 == T1:
                tmp["T"] = T0
            else:
                tmp["T0"] = T0
                tmp["T1"] = T1
            title1 = "!MolSSI trajectory 2.0 " + json.dumps(tmp, separators=(",", ":"))
            filename = f"@{self._id[-1]}+npt_state.trj"
            lines.append(
                f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
                f"                       {properties} &\n"
                "                       off 2 &\n"
                f"                       title1 '{title1}' &\n"
                f"                       title2 '{title2}' &\n"
                f"                       file {filename}"
            )
            self.description.append(
                __(
                    (
                        "The run will be {nsteps:,d} steps of dynamics "
                        "sampled every {nevery:n} steps."
                    ),
                    nsteps=nsteps,
                    nevery=nevery,
                    indent=7 * " ",
                )
            )

        # Handle trajectories
        tmp, ncomputes, ndumps, nfixes = self.trajectory_input(
            P, timestep, nsteps, ncomputes, ndumps, nfixes
        )
        lines.extend(tmp)

        if extras is not None and "shake" in extras:
            nfixes += 1
            lines.append(extras["shake"].format(nfixes))

        lines.append("")
        lines.append("run                 {}".format(nsteps))
        lines.append("")
        filename = f"@{self._id[-1]}+npt.dump"
        if configuration.periodicity == 0:
            lines.append(
                f"write_dump         all custom  {filename} id xu yu zu vx vy vz"
                " modify flush yes sort id"
            )
        else:
            lines.append(
                f"write_dump         all custom  {filename} id xsu ysu zsu vx vy vz"
                " modify flush yes sort id"
            )
        lines.append("")

        for fix in range(1, nfixes + 1):
            lines.append("unfix               {}".format(fix))
        lines.append("")

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }

    def get_pressure_text(self, _P):
        """Work out and return the pressure/stress part of the
        'fix npt' or 'fix berendsen' in LAMMPS
        """
        system_type = _P["system type"]
        Panneal = _P["Panneal"]

        if system_type == "fluid":
            P0 = lammps_step.to_lammps_units(_P["Pinitial"], quantity="pressure")
            Pdamp = lammps_step.to_lammps_units(_P["Pdamp"], quantity="time")
            if Panneal:
                P1 = lammps_step.to_lammps_units(_P["Pfinal"], quantity="pressure")
                return f" iso {P0:.3f} {P1:.3f} {Pdamp:.3f}"
            else:
                return f" iso {P0:.3f} {P0:.3f} {Pdamp:.3f}"

        use_pressure = "pressure" in _P["use_stress"]
        couple = _P["couple"]
        allow_shear = _P["allow shear"]

        if couple == "x, y and z":
            ptext = " couple xyz"
        elif couple == "x and y":
            ptext = " couple xy"
        elif couple == "x and z":
            ptext = " couple xz "
        elif couple == "y and z":
            ptext = " couple yz "
        else:
            ptext = " couple none"

        if use_pressure:
            P0 = lammps_step.to_lammps_units(_P["Pinitial"], quantity="pressure")
            if Panneal:
                P1 = lammps_step.to_lammps_units(_P["Pfinal"], quantity="pressure")
            else:
                P1 = P0
            Pdamp = lammps_step.to_lammps_units(_P["Pdamp"], quantity="time")

            P0 = round(P0, 3)
            P1 = round(P1, 3)
            Pdamp = round(Pdamp, 3)

            ptext += f" x {P0} {P1} {Pdamp} y {P0} {P1} {Pdamp} z {P0} {P1} {Pdamp}"
            if allow_shear:
                ptext += f" xy 0.0 0.0 {Pdamp} xz 0.0 0.0 {Pdamp} yz 0.0 0.0 {Pdamp}"
        else:
            Sxx0 = _P["Sxx,initial"]
            Syy0 = _P["Syy,initial"]
            Szz0 = _P["Szz,initial"]
            if Sxx0 != "fixed":
                Sxx0 = lammps_step.to_lammps_units(Sxx0, quantity="pressure")
            if Syy0 != "fixed":
                Syy0 = lammps_step.to_lammps_units(Syy0, quantity="pressure")
            if Szz0 != "fixed":
                Szz0 = lammps_step.to_lammps_units(Szz0, quantity="pressure")
            if Panneal:
                Sxx1 = _P["Sxx,final"]
                Syy1 = _P["Syy,final"]
                Szz1 = _P["Szz,final"]
                if Sxx1 != "fixed":
                    Sxx1 = lammps_step.to_lammps_units(Sxx1, quantity="pressure")
                if Syy1 != "fixed":
                    Syy1 = lammps_step.to_lammps_units(Syy1, quantity="pressure")
                if Szz1 != "fixed":
                    Szz1 = lammps_step.to_lammps_units(Szz1, quantity="pressure")
            else:
                Sxx1 = Sxx0
                Syy1 = Syy0
                Szz1 = Szz0
            Dxx = lammps_step.to_lammps_units(_P["Sxx damp"], quantity="pressure")
            Dyy = lammps_step.to_lammps_units(_P["Syy damp"], quantity="pressure")
            Dzz = lammps_step.to_lammps_units(_P["Szz damp"], quantity="pressure")

            if couple == "x, y and z":
                Syy0 = Szz0 = Sxx0
                Syy1 = Szz1 = Sxx1
                Dyy = Dzz = Dxx
            elif couple == "x and y":
                Syy0 = Sxx0
                Syy1 = Sxx1
                Dyy = Dxx
            elif couple == "x and z":
                Szz0 = Sxx0
                Szz1 = Sxx1
                Dzz = Dxx
            elif couple == "y and z":
                Szz0 = Syy0
                Szz1 = Szz1
                Dzz = Dyy

            # LAMMPS uses pressures = -stress, & also format nicely
            if Sxx0 != "fixed":
                ptext += f" x {-Sxx0:.3f} {-Sxx1:.3f} {Dxx:.3f} "
            if Syy0 != "fixed":
                ptext += f" y {-Syy0:.3f} {-Syy1:.3f} {Dyy:.3f} "
            if Szz0 != "fixed":
                ptext += f" z {-Szz0:.3f} {-Szz1:.3f} {Dzz:.3f} "

            if allow_shear:
                Syz0 = _P["Syz,initial"]
                Sxz0 = _P["Sxz,initial"]
                Sxy0 = _P["Sxy,initial"]
                if Syz0 != "fixed":
                    Syz0 = lammps_step.to_lammps_units(Syz0, quantity="pressure")
                if Sxz0 != "fixed":
                    Sxz0 = lammps_step.to_lammps_units(Sxz0, quantity="pressure")
                if Sxy0 != "fixed":
                    Sxy0 = lammps_step.to_lammps_units(Sxy0, quantity="pressure")
                if Panneal:
                    Syz1 = _P["Syz,final"]
                    Sxz1 = _P["Sxz,final"]
                    Sxy1 = _P["Sxy,final"]
                    if Syz1 != "fixed":
                        Syz1 = lammps_step.to_lammps_units(Syz1, quantity="pressure")
                    if Sxz1 != "fixed":
                        Sxz1 = lammps_step.to_lammps_units(Sxz1, quantity="pressure")
                    if Sxy1 != "fixed":
                        Sxy1 = lammps_step.to_lammps_units(Sxy1, quantity="pressure")
                else:
                    Syz1 = Syz0
                    Sxz1 = Sxz0
                    Sxy1 = Sxy0
                Dyz = lammps_step.to_lammps_units(_P["Syz damp"], quantity="pressure")
                Dxz = lammps_step.to_lammps_units(_P["Sxz damp"], quantity="pressure")
                Dxy = lammps_step.to_lammps_units(_P["Sxy damp"], quantity="pressure")

                # LAMMPS uses pressures = -stress, & also format nicely
                if Syz0 != "fixed":
                    ptext += f" yz {-Syz0:.3f} {-Syz1:.3f} {Dyz:.3f} "
                if Sxz0 != "fixed":
                    ptext += f" xz {-Sxz0:.3f} {-Sxz1:.3f} {Dxz:.3f} "
                if Sxy0 != "fixed":
                    ptext += f" xy {-Sxy0:.3f} {-Sxy1:.3f} {Dxy:.3f} "
        return ptext



# -*- coding: utf-8 -*-

"""A single-point energy in LAMMPS"""

from collections import Counter
import json
import logging
from math import isnan
from pathlib import Path
import pkg_resources
import textwrap
import traceback

import pandas
from tabulate import tabulate

import lammps_step
from lammps_step import LAMMPS, from_lammps_units
from molsystem import elements, Cell
import seamm
from seamm_util import units_class, Q_
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
printer = printing.getPrinter("lammps")


_subscript = {
    "0": "\N{SUBSCRIPT ZERO}",
    "1": "\N{SUBSCRIPT ONE}",
    "2": "\N{SUBSCRIPT TWO}",
    "3": "\N{SUBSCRIPT THREE}",
    "4": "\N{SUBSCRIPT FOUR}",
    "5": "\N{SUBSCRIPT FIVE}",
    "6": "\N{SUBSCRIPT SIX}",
    "7": "\N{SUBSCRIPT SEVEN}",
    "8": "\N{SUBSCRIPT EIGHT}",
    "9": "\N{SUBSCRIPT NINE}",
}


def subscript(n):
    """Return the number using Unicode subscript characters."""
    return "".join([_subscript[c] for c in str(n)])


one_half = "\N{VULGAR FRACTION ONE HALF}"
degree_sign = "\N{DEGREE SIGN}"
standard_state = {
    "H": f"{one_half}H{subscript(2)}(g)",
    "He": "He(g)",
    "Li": "Li(s)",
    "Be": "Be(s)",
    "B": "B(s)",
    "C": "C(s,gr)",
    "N": f"{one_half}N{subscript(2)}(g)",
    "O": f"{one_half}O{subscript(2)}(g)",
    "F": f"{one_half}F{subscript(2)}(g)",
    "Ne": "Ne(g)",
    "Na": "Na(s)",
    "Mg": "Mg(s)",
    "Al": "Al(s)",
    "Si": "Si(s)",
    "P": "P(s)",
    "S": "S(s)",
    "Cl": f"{one_half}Cl{subscript(2)}(g)",
    "Ar": "Ar(g)",
    "K": "K(s)",
    "Ca": "Ca(s)",
    "Sc": "Sc(s)",
    "Ti": "Ti(s)",
    "V": "V(s)",
    "Cr": "Cr(s)",
    "Mn": "Mn(s)",
    "Fe": "Fe(s)",
    "Co": "Co(s)",
    "Ni": "Ni(s)",
    "Cu": "Cu(s)",
    "Zn": "Zn(s)",
    "Ga": "Ga(s)",
    "Ge": "Ge(s)",
    "As": "As(s)",
    "Se": "Se(s)",
    "Br": f"{one_half}Br{subscript(2)}(l)",
    "Kr": "(g)",
}


class Energy(seamm.Node):
    """Handle a singlepoint energy calculation in LAMMPS"""

    def __init__(
        self,
        flowchart=None,
        title="Energy",
        extension=None,
        logger=logger,
    ):
        """Initialize the node"""

        logger.debug("Creating Energy {}".format(self))

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            logger=logger,
        )

        self._calculation = "energy"
        self._model = None
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.EnergyParameters()

        self.description = "A single point energy calculation"

    @property
    def model(self):
        """The name of the forcefield."""
        return self.parent.model

    @model.setter
    def model(self, value):
        self.parent.model = value

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def results(self):
        """The storage for the results in the main LAMMPS step."""
        return self.flowchart.parent._results

    @property
    def version(self):
        """The semantic version of this module."""
        return lammps_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return lammps_step.__git_revision__

    def description_text(self, P=None):
        """Create the text description of what this step will do."""

        text = "Single-point energy calculation."

        return self.header + "\n" + __(text, indent=4 * " ").__str__()

    def analyze(self, indent="", data={}, table=None, output=[], **kwargs):
        """Parse the output and generating the text output and store the
        data in variables for other stages to access
        """
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Get the initial structure
        _, initial_configuration = self.get_system_configuration()

        # Handle the new structure as needed
        system, configuration = self.get_system_configuration(P, model=self.model)

        # See if there are other results in json files
        filename = self.wd / "energy.json"
        if filename.exists():
            try:
                with filename.open("r") as fd:
                    tmp = json.load(fd)
            except Exception as e:
                printer.normal(f"Warning: error reading {filename}: {e}")
                logger.warning(f"Error with {filename}: {e}")
            else:
                data.update(tmp)

        # Get the stress vector
        stress = []
        if initial_configuration.periodicity != 0:
            for key in ("Sxx", "Syy", "Szz", "Syz", "Sxz", "Sxy"):
                if key in data:
                    value = Q_(data[key], data[key + ",units"]).m_as("atm")
                    stress.append(value)
                    data[key] = value
                    data[key + ",units"] = "atm"
        if len(stress) == 6:
            data["stress"] = stress
            data["stress,units"] = "atm"

        # Pressure
        if "P" in data:
            data["P"] = Q_(data["P"], data["P,units"]).m_as("atm")
            data["P,units"] = "atm"

        # Energies
        for key in ("Epe", "Etot"):
            if key in data:
                data[key] = Q_(data[key], data[key + ",units"]).m_as("kcal/mol")
                data[key + ",units"] = "kcal/mol"

        # Process the output to get the energy
        lines = iter(output)
        for line in lines:
            if "   Step" in line:
                tmp = line.split()
                try:
                    i = tmp.index("TotEng")
                except ValueError:
                    pass
                else:
                    line = next(lines)
                    tmp = line.split()
                    energy = from_lammps_units(float(tmp[i]), "kcal/mol").magnitude
                    data["energy"] = energy
                    data["energy,units"] = "kcal/mol"

                # Check for reaxff enthalpy offset
                if self.parent.ff_form() == "reaxff":
                    Eat = self.parent._atomic_energy_sum
                    if Eat != 0.0:
                        dHf = data["energy"] + Eat
                        data["DfH0_reax"] = dHf
                    data["energy,units"] = "kcal/mol"
            if line.startswith("Loop time of"):
                try:
                    tmp = line.split()
                    _time = round(float(tmp[3]), 2)
                    _procs = int(tmp[5])
                    _steps = int(tmp[8])
                    _natoms = int(tmp[11])
                    _type = self._calculation
                    self.results[f"t_{_type}"] = _time
                    self.results[f"np_{_type}"] = _procs
                    self.results[f"steps_{_type}"] = _steps
                    self.results[f"natoms_{_type}"] = _natoms
                except Exception as _e:
                    print(f"LAMMPS loop time: {_e}")

        # Calculate the enthalpy of formation, if possible
        tmp_text = self.calculate_enthalpy_of_formation(data)
        if tmp_text != "":
            path = self.wd / "Thermochemistry.txt"
            path.write_text(tmp_text)

        # Add static properties such as density for e.g NVE and NVT calculations
        if initial_configuration.periodicity == 3:
            if "density" not in data:
                data["density"] = initial_configuration.density
            if "a" not in data:
                data["a"] = initial_configuration.cell.a
            if "b" not in data:
                data["b"] = initial_configuration.cell.b
            if "c" not in data:
                data["c"] = initial_configuration.cell.c
            if "alpha" not in data:
                data["alpha"] = initial_configuration.cell.alpha
            if "beta" not in data:
                data["beta"] = initial_configuration.cell.beta
            if "gamma" not in data:
                data["gamma"] = initial_configuration.cell.gamma
            if "volume" not in data:
                data["volume"] = initial_configuration.cell.volume

        # Print the table of results
        if table is None:
            table = {
                "Property": [],
                "Value": [],
                "Units": [],
            }
        for key in ("DfH0_reax", "DfE0", "energy"):
            if key in data:
                table["Property"].append(key)
                table["Value"].append(f"{data[key]:.2f}")
                table["Units"].append("kcal/mol")
        for key in ("a", "b", "c"):
            if key in data:
                table["Property"].append(key)
                table["Value"].append(f"{data[key]:.3f}")
                table["Units"].append("Å")
        for key, char in zip(("alpha", "beta", "gamma"), ("𝞪", "𝞫", "𝞬")):
            if key in data:
                table["Property"].append(char)
                table["Value"].append(f"{data[key]:.3f}")
                table["Units"].append("°")
        for key in ("P", "Sxx", "Syy", "Szz", "Syz", "Sxz", "Sxy"):
            if key in data:
                table["Property"].append(key)
                table["Value"].append(f"{data[key]:.2f}")
                table["Units"].append("atm")
        for key in ("density",):
            if key in data:
                table["Property"].append(key)
                table["Value"].append(f"{data[key]:.3f}")
                table["Units"].append("g/mL")
        for key in ("volume",):
            if key in data:
                table["Property"].append(key)
                table["Value"].append(f"{data[key]:.1f}")
                table["Units"].append("Å^3")

        tmp = tabulate(
            table,
            headers="keys",
            tablefmt="rounded_outline",
            colalign=("center", "decimal", "left"),
            disable_numparse=True,
        )
        length = len(tmp.splitlines()[0])
        text_lines = []
        text_lines.append("Single Point Energy".center(length))
        text_lines.append(tmp)
        printer.normal(textwrap.indent("\n".join(text_lines), self.indent + 7 * " "))
        printer.normal("")

        # Read the dump file and get the structure
        xyz = None
        cell = None
        gradients = None
        velocities = None
        dump_file = self.wd / "energy.dump"
        if dump_file.exists:
            try:
                tmp = dump_file.read_text().splitlines()
                results = self._parse_dump_frame(tmp)
                if "xyz" in results:
                    xyz = results["xyz"]
                    fractional = False
                elif "abc" in results:
                    xyz = results["abc"]
                    fractional = True
                else:
                    xyz = None
                cell = results["cell"] if "cell" in results else None
                gradients = results["gradients"] if "gradients" in results else None
                velocities = results["velocities"] if "velocities" in results else None
                if gradients is not None:
                    data["gradients"] = gradients
                if velocities is not None:
                    data["velocities"] = velocities
            except Exception as e:
                printer.normal("Warning: unable to read the LAMMPS dumpfile")
                logger.warning(f"The was a problem reading the LAMMPS dumpfile: {e}")
                logger.warning(traceback.format_exc())
        else:
            printer.normal("Warning: there is no 'dump' file from LAMMPS")

        if configuration is not None:
            if cell is not None:
                configuration.cell.parameters = cell
            if xyz is not None:
                configuration.atoms.set_coordinates(xyz, fractionals=fractional)
            if gradients is not None:
                # LAMMPS only has Cartesian gradients
                configuration.atoms.set_gradients(gradients, fractionals=False)
            if velocities is not None:
                # LAMMPS only has Cartesian velocities
                configuration.atoms.set_velocities(velocities, fractionals=False)

            # And the name of the configuration.
            text = seamm.standard_parameters.set_names(
                system,
                configuration,
                P,
                _first=True,
                model=self.model,
            )
            printer.normal(__(text, **data, indent=self.indent + 4 * " "))
            printer.normal("")

        # Put any requested results into variables or tables
        self.store_results(
            configuration=configuration,
            data=data | self.results,
            create_tables=self.parameters["create tables"].get(),
            printer=printer,
        )

    def calculate_enthalpy_of_formation(self, data):
        """Calculate the enthalpy of formation from the results of a calculation.

        This uses tabulated values of the enthalpy of formation of the atoms for
        the elements and tabulated energies calculated for atoms with the current
        method.

        Parameters
        ----------
        data : dict
            The results of the calculation.
        """
        J_to_cal = Q_("J").m_as("cal")

        # Read the tabulated values from either user or data directory
        personal_file = Path("~/.seamm.d/data/element_energies.csv").expanduser()
        if personal_file.exists():
            personal_table = pandas.read_csv(personal_file, index_col=False)
        else:
            personal_table = None

        path = Path(pkg_resources.resource_filename(__name__, "data/"))
        csv_file = path / "element_energies.csv"
        table = pandas.read_csv(csv_file, index_col=False)

        self.logger.debug(f"self.model = {self.model}")

        # Check if have the data
        atom_formation_energy = None
        atom_energy = None
        column = self.model

        self.logger.debug(f"Looking for '{column}'")

        # atom_formation_energy is the energy of formation of the standard state,
        # per atom.
        # atom_energy is the calculated energy of the atom, which defaults to zero
        column2 = column + " atom energy"
        if personal_table is not None and column in personal_table.columns:
            atom_formation_energy = (J_to_cal * personal_table[column]).to_list()
            if column2 in personal_table.columns:
                atom_energy = (J_to_cal * personal_table[column2]).to_list()
        elif column in table.columns:
            atom_formation_energy = (J_to_cal * table[column]).to_list()
            if column2 in table.columns:
                atom_energy = (J_to_cal * table[column2]).to_list()

        if atom_formation_energy is None:
            # Not found!
            return f"There are no tabulated atom energies for {column}"

        # Assume an offset energy -- the energy of an isolated atom -- is zero if not
        # tabulated
        if atom_energy is None:
            atom_energy = [0.0] * len(atom_formation_energy)

        DfH0gas = None
        references = None
        term_symbols = None
        if personal_table is not None and "ΔfH°gas" in personal_table.columns:
            DfH0gas = personal_table["ΔfH°gas"].to_list()
            if "Reference" in personal_table.columns:
                references = personal_table["Reference"].to_list()
            if "Term Symbol" in personal_table.columns:
                term_symbols = personal_table["Term Symbols"].to_list()
        elif "ΔfH°gas" in table.columns:
            DfH0gas = table["ΔfH°gas"].to_list()
            if "Reference" in table.columns:
                references = table["Reference"].to_list()
            if "Term Symbol" in table.columns:
                term_symbols = table["Term Symbol"].to_list()

        # Get the atomic numbers and counts
        _, configuration = self.get_system_configuration(None)
        counts = Counter(configuration.atoms.atomic_numbers)

        # Get the Hill formula as a list
        symbols = sorted(elements.to_symbols(counts.keys()))
        composition = []
        if "C" in symbols:
            composition.append((6, "C", counts[6]))
            symbols.remove("C")
            if "H" in symbols:
                composition.append((1, "H", counts[1]))
                symbols.remove("H")

        for symbol in symbols:
            atno = elements.symbol_to_atno[symbol]
            composition.append((atno, symbol, counts[atno]))

        # And the reactions. First, for atomization energy
        middot = "\N{MIDDLE DOT}"
        lDelta = "\N{GREEK CAPITAL LETTER DELTA}"
        formula = ""
        tmp = []
        for atno, symbol, count in composition:
            if count == 1:
                formula += symbol
                tmp.append(f"{symbol}(g)")
            else:
                formula += f"{symbol}{subscript(count)}"
                tmp.append(f"{count}{middot}{symbol}(g)")
        gas_atoms = " + ".join(tmp)
        tmp = []
        for atno, symbol, count in composition:
            if count == 1:
                tmp.append(standard_state[symbol])
            else:
                tmp.append(f"{count}{middot}{standard_state[symbol]}")
        standard_elements = " + ".join(tmp)

        # The energy - any offsets is the negative of the atomization energy
        name = "Formula: " + formula
        try:
            name = configuration.PC_iupac_name(fallback=name)
        except Exception:
            pass

        if name is None:
            name = "Formula: " + formula

        text = f"Thermochemistry of {name} with {column}\n\n"
        text += "Atomization Energy\n"
        text += "------------------\n"
        text += textwrap.fill(
            f"The atomization energy,  {lDelta}atE{degree_sign}, is the energy to break"
            " all the bonds in the system, separating the atoms from each other."
        )
        text += f"\n\n    {formula} --> {gas_atoms}\n\n"
        text += textwrap.fill(
            "The following table shows in detail the calculation. The first line is "
            "the system and its calculated energy. The next lines are the energies "
            "of each type of atom in the system. These have been tabulated by running "
            "calculations on each atom, and are included in the SEAMM release. "
            "The line give the formation energy from atoms in kcal/mol.",
        )
        text += "\n\n"
        table = {
            "System": [],
            "Term": [],
            "Value": [],
            "Units": [],
        }

        if "Epe" in data:
            E = data["Epe"]
        elif "energy" in data:
            E = data["energy"]
        else:
            return "The energy is not in results from the calculation!"

        Eatoms = 0.0
        Ef0 = 0.0
        for atno, symbol, count in composition:
            Eatom = atom_energy[atno - 1]
            if isnan(Eatom):
                # Don't have the data for this element
                return f"Do not have tabulated atom energies for {symbol} in {column}"
            Eatoms += count * Eatom
            table["System"].append(f"{symbol}(g)")
            table["Term"].append(f"{count} * {Eatom:.2f}")
            table["Value"].append(f"{count * Eatom:.2f}")
            table["Units"].append("")

            Ef0 += count * atom_formation_energy[atno - 1]

        data["DfE0"] = E - Ef0

        table["Units"][0] = "kcal/mol"

        table["System"].append("^")
        table["Term"].append("-")
        table["Value"].append("-")
        table["Units"].append("")

        table["System"].append(formula)
        table["Term"].append(f"{-E:.2f}")
        table["Value"].append(f"{-E:.2f}")
        table["Units"].append("kcal/mol")

        data["E atomization"] = Eatoms - E

        table["System"].append("")
        table["Term"].append("")
        table["Value"].append("=")
        table["Units"].append("")

        result = f'{data["E atomization"]:.2f}'
        table["System"].append(f"{lDelta}atE")
        table["Term"].append("")
        table["Value"].append(result)
        table["Units"].append("kcal/mol")

        factor = Q_("kcal/mol").m_as("kJ/mol")
        table["System"].append("")
        table["Term"].append("")
        table["Value"].append(f'{factor * data["E atomization"]:.2f}')
        table["Units"].append("kJ/mol")

        tmp = tabulate(
            table,
            headers="keys",
            tablefmt="rounded_outline",
            colalign=("center", "center", "decimal", "center"),
            disable_numparse=True,
        )
        length = len(tmp.splitlines()[0])
        text_lines = []
        text_lines.append(f"Atomization Energy for {formula}".center(length))
        text_lines.append(tmp)
        text += textwrap.indent("\n".join(text_lines), 4 * " ")

        if "H" not in data:
            text += "\n\n"
            text += "Cannot calculate enthalpy of formation without the enthalpy"
            return text
        if DfH0gas is None:
            text += "\n\n"
            text += "Cannot calculate enthalpy of formation without the tabulated\n"
            text += "atomization enthalpies of the elements."
            return text

        # Atomization enthalpy of the elements, experimental
        table = {
            "System": [],
            "Term": [],
            "Value": [],
            "Units": [],
            "Reference": [],
        }

        E = data["energy"]

        DfH_at = 0.0
        refno = 1
        for atno, symbol, count in composition:
            DfH_atom = DfH0gas[atno - 1]
            DfH_at += count * DfH_atom
            tmp = Q_(DfH_atom, "kcal/mol").m_as("E_h")
            table["System"].append(f"{symbol}(g)")
            if count == 1:
                table["Term"].append(f"{tmp:.6f}")
            else:
                table["Term"].append(f"{count} * {tmp:.6f}")
            table["Value"].append(f"{count * tmp:.6f}")
            table["Units"].append("")
            refno += 1
            table["Reference"].append(refno)

        table["Units"][0] = "E_h"

        table["System"].append("^")
        table["Term"].append("-")
        table["Value"].append("-")
        table["Units"].append("")
        table["Reference"].append("")

        table["System"].append(standard_elements)
        table["Term"].append("")
        table["Value"].append("0.0")
        table["Units"].append("E_h")
        table["Reference"].append("")

        table["System"].append("")
        table["Term"].append("")
        table["Value"].append("=")
        table["Units"].append("")
        table["Reference"].append("")

        result = f'{Q_(DfH_at, "kcal/mol").m_as("E_h"):.6f}'
        table["System"].append(f"{lDelta}atH{degree_sign}")
        table["Term"].append("")
        table["Value"].append(result)
        table["Units"].append("E_h")
        table["Reference"].append("")

        table["System"].append("")
        table["Term"].append("")
        table["Value"].append(f"{DfH_at:.2f}")
        table["Units"].append("kcal/mol")
        table["Reference"].append("")

        tmp = tabulate(
            table,
            headers="keys",
            tablefmt="rounded_outline",
            colalign=("center", "center", "decimal", "center", "center"),
            disable_numparse=True,
        )
        length = len(tmp.splitlines()[0])
        text_lines = []
        text_lines.append(
            "Atomization enthalpy of the elements (experimental)".center(length)
        )
        text_lines.append(tmp)

        text += "\n\n"
        text += "Enthalpy of Formation\n"
        text += "---------------------\n"
        text += textwrap.fill(
            f"The enthalpy of formation, {lDelta}fHº, is the enthalpy of creating the "
            "molecule from the elements in their standard state:"
        )
        text += f"\n\n   {standard_elements} --> {formula} (1)\n\n"
        text += textwrap.fill(
            "The standard state of the element, denoted by the superscript º,"
            " is its form at 298.15 K and 1 atm pressure, e.g. graphite for carbon, "
            "H2 gas for hydrogen, etc."
        )
        text += "\n\n"
        text += textwrap.fill(
            "Since it is not easy to calculate the enthalpy of e.g. graphite we will "
            "use two sequential reactions that are equivalent. First, we will create "
            "gas phase atoms from the elements:"
        )
        text += f"\n\n    {standard_elements} --> {gas_atoms} (2)\n\n"
        text += textwrap.fill(
            "This will use the experimental values of the enthalpy of formation of the "
            "atoms in the gas phase to calculate the enthalpy of this reaction. "
            "Then we react the atoms to get the desired system:"
        )
        text += f"\n\n    {gas_atoms} --> {formula} (3)\n\n"
        text += textwrap.fill(
            "Note that this is reverse of the atomization reaction, so "
            f"{lDelta}H = -{lDelta}atH."
        )
        text += "\n\n"
        text += textwrap.fill(
            "First we calculate the enthalpy of the atomization of the elements in "
            "their standard state, using tabulated experimental values:"
        )
        text += "\n\n"
        text += textwrap.indent("\n".join(text_lines), 4 * " ")

        # And the calculated atomization enthalpy
        table = {
            "System": [],
            "Term": [],
            "Value": [],
            "Units": [],
        }

        Hatoms = 0.0
        dH = Q_(6.197, "kcal/mol").m_as("E_h")
        for atno, symbol, count in composition:
            Eatom = atom_formation_energy[atno - 1]
            # 6.197 is the H298-H0 for an atom
            Hatoms += count * (Eatom + 6.197)

            table["System"].append(f"{symbol}(g)")
            if count == 1:
                table["Term"].append(f"{-Eatom:.2f} + {dH:.2f}")
            else:
                table["Term"].append(f"{count} * ({-Eatom:.2f} + {dH:.2f})")
            table["Value"].append(f"{-count * (Eatom + dH):.2f}")
            table["Units"].append("")

        table["System"].append("^")
        table["Term"].append("-")
        table["Value"].append("-")
        table["Units"].append("")

        H = data["H"]

        table["System"].append(formula)
        table["Term"].append(f"{H:.2f}")
        table["Value"].append("")
        table["Units"].append("kcal/mol")

        data["H atomization"] = Hatoms - Q_(H, "E_h").m_as("kcal/mol")
        data["DfH0"] = DfH_at - data["H atomization"]
        table["System"].append("")
        table["Term"].append("")
        table["Value"].append("=")
        table["Units"].append("")

        table["System"].append("")
        table["Term"].append("")
        table["Value"].append(f'{data["H atomization"]:.2f}')
        table["Units"].append("kcal/mol")

        tmp = tabulate(
            table,
            headers="keys",
            tablefmt="rounded_outline",
            colalign=("center", "center", "decimal", "center"),
            disable_numparse=True,
        )
        length = len(tmp.splitlines()[0])
        text_lines = []
        text_lines.append("Atomization Enthalpy (calculated)".center(length))
        text_lines.append(tmp)
        text += "\n\n"

        text += textwrap.fill(
            "Next we calculate the atomization enthalpy of the system. We have the "
            "calculated enthalpy of the system, but need the enthalpy of gas phase "
            f"atoms at the standard state (25{degree_sign}C, 1 atm). The tabulated "
            "energies for the atoms, used above, are identical to H0 for an atom. "
            "We will add H298 - H0 to each atom, which [1] is 5/2RT = 0.002360 E_h"
        )
        text += "\n\n"
        text += textwrap.indent("\n".join(text_lines), 4 * " ")
        text += "\n\n"
        text += textwrap.fill(
            "The enthalpy change for reaction (3) is the negative of this atomization"
            " enthalpy. Putting the two reactions together with the negative for Rxn 3:"
        )
        text += "\n\n"
        text += f"{lDelta}fH{degree_sign} = {lDelta}H(rxn 2) - {lDelta}H(rxn 3)\n"
        text += f"     = {DfH_at:.2f} - {data['H atomization']:.2f}\n"
        text += f"     = {DfH_at - data['H atomization']:.2f} kcal/mol\n"

        text += "\n\n"
        text += "References\n"
        text += "----------\n"
        text += "1. https://en.wikipedia.org/wiki/Monatomic_gas\n"
        refno = 1
        for atno, symbol, count in composition:
            refno += 1
            text += f"{refno}. {lDelta}fH{degree_sign} = {DfH0gas[atno - 1]} kcal/mol"
            if term_symbols is not None:
                text += f" for {term_symbols[atno - 1]} {symbol}"
            else:
                text += f" for {symbol}"
            if references is not None:
                text += f" from {references[atno-1]}\n"

        return text

    def get_input(self, extras=None):
        """Get the input for an energy calculation for LAMMPS"""

        P = self.parameters.values_to_dict()

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        _, configuration = self.get_system_configuration()

        self.description = []
        self.description.append(__(self.description_text(PP), **PP, indent=4 * " "))

        lines = []

        filename = f"@{self._id[-1]}+forces.dump"
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")
        lines.append(f"dump                1 all custom 1 {filename} id fx fy fz")
        lines.append("run                 0")
        lines.append("undump              1")

        if configuration.periodicity != 0:
            # Write out the stress
            filename = f"@{self._id[-1]}+energy.json"
            units = lammps_step.lammps_units("pressure")
            eunits = lammps_step.lammps_units("energy")
            lines.append(
                'print               """{\n'
                '    "P": $(v_press:%.3f),\n'
                f'    "P,units": "{units}",\n'
                '    "Sxx": $(v_sxx:%.3f),\n'
                f'    "Sxx,units": "{units}",\n'
                '    "Syy": $(v_syy:%.3f),\n'
                f'    "Syy,units": "{units}",\n'
                '    "Szz": $(v_szz:%.3f),\n'
                f'    "Szz,units": "{units}",\n'
                '    "Syz": $(v_syz:%.3f),\n'
                f'    "Syz,units": "{units}",\n'
                '    "Sxz": $(v_sxz:%.3f),\n'
                f'    "Sxz,units": "{units}",\n'
                '    "Sxy": $(v_sxy:%.3f),\n'
                f'    "Sxy,units": "{units}",\n'
                '    "Etot": $(v_etotal:%.4f),\n'
                f'    "Etot,units": "{eunits}",\n'
                '    "Epe": $(v_etotal:%.4f),\n'
                f'    "Epe,units": "{eunits}"\n'
                "}"
                f'""" file {filename}'
            )
        filename = f"@{self._id[-1]}+energy.dump"
        lines.append(
            f"write_dump          all custom  {filename} id xu yu zu fx fy fz"
            " modify flush yes sort id"
        )

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }

    def _parse_dump_frame(self, lines, line_no=0):
        """Process a single frame of a LAMMPS dump file.

        Parameters
        ----------
        lines : [str]
            The lines of the section.

        lineno : int
            The count of lines before this frame, used for error messages

        Returns
        -------
        dict(str, any)
            The results keyed by the item name.
            The units are Å, kJ/mol, and fs
        """
        results = {}
        it = iter(lines)

        def next_line(it, line_no):
            return (next(it).strip(), line_no + 1)

        line, line_no = next_line(it, line_no)

        while True:
            try:
                if not line.startswith("ITEM:"):
                    raise ValueError(
                        f"Error reading dump file: line {line_no} should be 'ITEM: ...'"
                        f"\n\t'{line}'"
                    )
                section = line[6:].strip()
                self.logger.debug("   section = " + section)

                if section == "TIMESTEP":
                    line, line_no = next_line(it, line_no)
                    results["timestep"] = int(line)
                    line, line_no = next_line(it, line_no)
                    continue
                elif section == "NUMBER OF ATOMS":
                    line, line_no = next_line(it, line_no)
                    results["n_atoms"] = n_atoms = int(line)
                    line, line_no = next_line(it, line_no)
                    continue
                elif "BOX BOUNDS" in section:
                    factor = from_lammps_units(1, "Å").magnitude
                    if len(section.split()) == 8:
                        line, line_no = next_line(it, line_no)
                        xlo_bound, xhi_bound, xy = line.split()
                        line, line_no = next_line(it, line_no)
                        ylo_bound, yhi_bound, xz = line.split()
                        line, line_no = next_line(it, line_no)
                        zlo, zhi, yz = line.split()

                        xlo_bound = float(xlo_bound) * factor
                        xhi_bound = float(xhi_bound) * factor
                        ylo_bound = float(ylo_bound) * factor
                        yhi_bound = float(yhi_bound) * factor
                        zlo = float(zlo) * factor
                        zhi = float(zhi) * factor
                        xy = float(xy) * factor
                        xz = float(xz) * factor
                        yz = float(yz) * factor

                        xlo = xlo_bound - min(0.0, xy, xz, xy + xz)
                        xhi = xhi_bound - max(0.0, xy, xz, xy + xz)
                        ylo = ylo_bound - min(0.0, yz)
                        yhi = yhi_bound - max(0.0, yz)
                        cell = LAMMPS.box_to_cell(
                            xhi - xlo, yhi - ylo, zhi - zlo, xy, xz, yz
                        )
                        tmp = Cell(*cell)
                        lattice = []
                        for v in tmp.vectors():
                            lattice.extend(v)
                    else:
                        line, line_no = next_line(it, line_no)
                        xlo, xhi = line.split()
                        line, line_no = next_line(it, line_no)
                        ylo, yhi = line.split()
                        line, line_no = next_line(it, line_no)
                        zlo, zhi = line.split()

                        xlo = float(xlo) * factor
                        xhi = float(xhi) * factor
                        ylo = float(ylo) * factor
                        yhi = float(yhi) * factor
                        zlo = float(zlo) * factor
                        zhi = float(zhi) * factor

                        lattice = (
                            xhi - xlo,
                            0.0,
                            0.0,
                            0.0,
                            yhi - ylo,
                            0.0,
                            0.0,
                            0.0,
                            zhi - zlo,
                        )
                        cell = (xhi - xlo, yhi - ylo, zhi - zlo, 90, 90, 90)
                    results["cell"] = cell
                    results["lattice"] = lattice
                    line, line_no = next_line(it, line_no)
                elif "ATOMS" in section:
                    xyz = []
                    f = []
                    v = []
                    keys = section.split()[1:]

                    # Coordinates
                    ixyz = None
                    if "x" in keys and "y" in keys and "z" in keys:
                        ixyz = (keys.index("x"), keys.index("y"), keys.index("z"))
                        xyz_type = "cartesian"
                        fxyz = from_lammps_units(1, "Å").magnitude
                    elif "xu" in keys and "yu" in keys and "zu" in keys:
                        ixyz = (keys.index("xu"), keys.index("yu"), keys.index("zu"))
                        xyz_type = "cartesian"
                        fxyz = from_lammps_units(1, "Å").magnitude
                    elif "xs" in keys and "ys" in keys and "zs" in keys:
                        ixyz = (keys.index("xs"), keys.index("ys"), keys.index("zs"))
                        xyz_type = "fractional"
                        fxyz = 1
                    elif "xsu" in keys and "ysu" in keys and "zsu" in keys:
                        ixyz = (keys.index("xsu"), keys.index("ysu"), keys.index("zsu"))
                        xyz_type = "fractional"
                        fxyz = 1

                    # Forces
                    fi = None
                    if "fx" in keys and "fy" in keys and "fz" in keys:
                        fi = (keys.index("fx"), keys.index("fy"), keys.index("fz"))
                        ff = -from_lammps_units(1, "kJ/mol/Å").magnitude

                    # Velocities
                    vi = None
                    if "vx" in keys and "vy" in keys and "vz" in keys:
                        vi = (keys.index("vx"), keys.index("vy"), keys.index("vz"))
                        fv = from_lammps_units(1, "Å/fs").magnitude

                    for i in range(n_atoms):
                        line, line_no = next_line(it, line_no)
                        values = line.split()
                        if ixyz is not None:
                            xyz.append(
                                [
                                    fxyz * float(values[ixyz[0]]),
                                    fxyz * float(values[ixyz[1]]),
                                    fxyz * float(values[ixyz[2]]),
                                ]
                            )
                        if fi is not None:
                            f.append(
                                [
                                    ff * float(values[fi[0]]),
                                    ff * float(values[fi[1]]),
                                    ff * float(values[fi[2]]),
                                ]
                            )
                        if vi is not None:
                            v.append(
                                [
                                    fv * float(values[vi[0]]),
                                    fv * float(values[vi[1]]),
                                    fv * float(values[vi[2]]),
                                ]
                            )
                    if ixyz is not None:
                        if xyz_type == "cartesian":
                            results["xyz"] = xyz
                        elif xyz_type == "fractional":
                            results["abc"] = xyz
                    if fi is not None:
                        results["gradients"] = f
                    if vi is not None:
                        results["velocities"] = v

                    line, line_no = next_line(it, line_no)
                else:
                    raise ValueError(
                        f"Don't recognize section {section} in the dump file"
                    )
            except StopIteration:
                break

        return results



# -*- coding: utf-8 -*-

"""The graphical part of a LAMMPS velocities step"""

import tkinter as tk

import seamm
import seamm_widgets as sw


class TkVelocities(seamm.TkNode):
    def __init__(
        self, tk_flowchart=None, node=None, canvas=None, x=None, y=None, w=200, h=50
    ):
        """Initialize a node

        Keyword arguments:
        """

        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

    def right_click(self, event):
        """Probably need to add our dialog..."""

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def create_dialog(self):
        """Create the dialog!"""

        frame = super().create_dialog("Edit LAMMPS Velocity Step")

        # Create all the widgets
        P = self.node.parameters
        for key in P:
            self[key] = P[key].widget(frame)

        self["method"].combobox.bind("<<ComboboxSelected>>", self.reset_dialog)
        self["method"].combobox.bind("<Return>", self.reset_dialog)
        self["method"].combobox.bind("<FocusOut>", self.reset_dialog)

        self.reset_dialog()

    def reset_dialog(self, widget=None):
        """Lay out the widgets given the current state"""
        method = self["method"].get()

        frame = self["frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        row = 0
        widgets = []
        for key in ("method", "T", "remove_momentum"):
            self[key].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self[key])
            row += 1

        if "scaling" not in method:
            self["seed"].grid(row=row, column=0, sticky=tk.EW)
            widgets.append(self["seed"])
            row += 1

            sw.align_labels(widgets, sticky=tk.E)



# -*- coding: utf-8 -*-

"""Control parameters for NVT (canonical) dynamics"""

import lammps_step
import logging

logger = logging.getLogger(__name__)


class NVT_Parameters(lammps_step.NVE_Parameters):
    """The control parameters for NVT dynamics in LAMMPS"""

    parameters = {
        "thermostat": {
            "default": "Nose-Hoover",
            "kind": "string",
            "format_string": "s",
            "enumeration": (
                "Nose-Hoover",
                "Berendsen",
                "canonical sampling, velocity rescaling (csvr)",
                "canonical sampling, langevin dynamics (csld)",
                "velocity rescaling",
                "Langevin",
            ),
            "description": "Thermostat:",
            "help_text": ("The thermostat used to control the " "temperature."),
        },
        "T0": {
            "default": 298.15,
            "kind": "float",
            "default_units": "K",
            "format_string": ".2f",
            "description": "Temperature:",
            "help_text": (
                "The temperature, or initial temperature for " "simulated annealing."
            ),
        },
        "T1": {
            "default": 298.15,
            "kind": "float",
            "default_units": "K",
            "format_string": ".2f",
            "description": "Final temperature:",
            "help_text": "The final temperature for simulated annealing.",
        },
        "Tdamp": {
            "default": 100.0,
            "kind": "float",
            "default_units": "fs",
            "format_string": ".1f",
            "description": "Damping time:",
            "help_text": "The damping time constant for thermostat",
        },
        "Tchain": {
            "default": 3,
            "kind": "integer",
            "default_units": None,
            "format_string": "d",
            "description": "Thermostat chain:",
            "help_text": "The number of thermostats in the chain.",
        },
        "Tloop": {
            "default": 1,
            "kind": "integer",
            "default_units": None,
            "format_string": "d",
            "description": "Thermostat iterations:",
            "help_text": ("The number of sub-iterations for the thermostat."),
        },
        "drag": {
            "default": 0.0,
            "kind": "float",
            "default_units": None,
            "format_string": ".1f",
            "description": "Drag:",
            "help_text": (
                "The amount of drag to apply on the thermostat "
                "and barostat if the pressure is controlled. "
                "Typically a value of 0.2 - 2.0 is appropriate."
            ),
        },
        "seed": {
            "default": "random",
            "kind": "integer",
            "default_units": None,
            "format_string": "",
            "enumeration": ("random",),
            "description": "Random seed:",
            "help_text": (
                "The seed for the random number generator."
                "'random' means to generate a random integer "
                "as the seed."
            ),
        },
        "frequency": {
            "default": 100.0,
            "kind": "float",
            "default_units": "fs",
            "format_string": ".1f",
            "description": "Rescaling frequency:",
            "help_text": (
                "The frequency (in time units) for the thermostat " "to operate."
            ),
        },
        "window": {
            "default": 20.0,
            "kind": "float",
            "default_units": "K",
            "format_string": ".1f",
            "description": "Temperature window:",
            "help_text": (
                "The temperature window to use when rescaling."
                "When the instantaneous temperature is outside "
                "the temperature window, the velocities will be "
                "rescaled."
            ),
        },
        "fraction": {
            "default": 1.0,
            "kind": "float",
            "default_units": None,
            "format_string": ".1f",
            "description": "Fraction to scale:",
            "help_text": (
                "The fraction of the temperature difference "
                "to correct by scaling the velocities."
            ),
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(defaults={**NVT_Parameters.parameters, **defaults}, data=data)



# -*- coding: utf-8 -*-

"""Set the velocities on the atoms

ToDo:
    * This does not support groups other than "all" yet. Once it does, setting
        specific velocities would also be useful.
    * Gaussian or normal distributions...
"""

import lammps_step
import logging
import seamm
from seamm_util import units_class
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import random

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("lammps")


class Velocities(seamm.Node):
    def __init__(self, flowchart=None, title="Velocities", extension=None):
        """Initialize the node"""

        logger.debug("Creating Velocities {}".format(self))

        super().__init__(flowchart=flowchart, title=title, extension=extension)

        self.description = "Set the initial velocities on the atoms"
        self.parameters = lammps_step.VelocitiesParameters()

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}".format(".".join(str(e) for e in self._id), self.title)

    @property
    def version(self):
        """The semantic version of this module."""
        return lammps_step.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return lammps_step.__git_revision__

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """

        if not P:
            P = self.parameters.values_to_dict()

        text = "Set the velocities to give a temperature {T} " "by {method}."

        if self.is_expr(P["remove_momentum"]):
            text += (
                " Whether to remove translational or rotational "
                "momentum will be determined at runtime by "
                "'{remove_momentum}'"
            )
        else:
            text += " LAMMPS will {remove_momentum}"

        if P["method"] != "scaling current velocities":
            if P["seed"] == "random":
                text += " The random number generator will be initialized randomly."
            else:
                text += (
                    " The random number generator will be initialized "
                    "with the seed '{seed}'."
                )

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def get_input(self, extras=None):
        """Get the input for setting the velocities in LAMMPS"""
        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        # Fix variables that need attention
        if "default" in P["remove_momentum"]:
            system_db = self.get_variable("_system_db")
            configuration = system_db.system.configuration
            if configuration.periodicity == 3:
                P["remove_momentum"] = (
                    "remove translational but not rotational momentum"
                )
            else:
                P["remove_momentum"] = (
                    "remove both translational and rotational momentum"
                )
        if P["seed"] == "random":
            P["seed"] = int(random.random() * 2**31)

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description = [str(__(self.description_text(PP), **PP, indent=4 * " "))]

        # Get the input lines
        lines = []
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")

        if P["remove_momentum"] == "remove translational but not rotational momentum":
            remove_translations = "yes"
            remove_rotations = "no"
        elif P["remove_momentum"] == "remove rotational but not translational momentum":
            remove_translations = "no"
            remove_rotations = "yes"
        elif P["remove_momentum"] == (
            "remove both translational and rotational momentum"
        ):
            remove_translations = "yes"
            remove_rotations = "yes"
        elif P["remove_momentum"] == (
            "remove neither translational nor rotational momentum"
        ):
            remove_translations = "no"
            remove_rotations = "no"
        else:
            raise RuntimeError(
                "Don't recognize 'remove_momentum' of '{}'".format(P["remove_momentum"])
            )

        T = P["T"].to("K").magnitude

        if "random" in P["method"]:
            lines.append(
                "velocity            all create {} {} mom {} rot {}".format(
                    T, P["seed"], remove_translations, remove_rotations
                )
            )
        elif "scaling" in P["method"]:
            lines.append(
                "velocity            all scale {} mom {} rot {}".format(
                    T, remove_translations, remove_rotations
                )
            )
        else:
            raise RuntimeError(
                "Velocity method '{}' not supported yet".format(P["method"])
            )

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }



# -*- coding: utf-8 -*-

"""The graphical part of a LAMMPS Energy step"""

import lammps_step
import logging
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk

logger = logging.getLogger(__name__)


class TkNVE(lammps_step.TkEnergy):
    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
        my_logger=logger,
    ):
        """Initialize a node

        Keyword arguments:
        """

        # Metadata for the properties
        self.property_metadata = {}

        super().__init__(
            tk_flowchart=tk_flowchart,
            node=node,
            canvas=canvas,
            x=x,
            y=y,
            w=w,
            h=h,
            my_logger=my_logger,
        )

    def create_dialog(self, title="Edit NVE dynamics parameters"):
        """Create the dialog!"""

        # Let parent classes do their thing.
        frame = super().create_dialog(title=title)

        # Shortcut for parameters
        P = self.node.parameters

        # Add a tab for the trajectory setup
        notebook = self["notebook"]
        tframe = ttk.Frame(notebook)
        self["trajectory frame"] = tframe
        notebook.insert(
            self["results frame"], tframe, text="Trajectories", sticky=tk.NSEW
        )
        for key in (
            "trajectory",
            "trajectory save",
            "trajectory export",
            "atomic positions",
            "com positions",
            "atomic velocities",
            "com velocities",
            "heat flux",
            "shear stress",
        ):
            if title == "Heat Flux" and (key == "heat flux" or "centroid" in key):
                continue
            self[key] = P[key].widget(tframe)
            self[key].bind("<<ComboboxSelected>>", self.reset_trajectory_frame)
            self[key].bind("<Return>", self.reset_trajectory_frame)
            self[key].bind("<FocusOut>", self.reset_trajectory_frame)

        for key in (
            "trajectory rate",
            "trajectory number of samples",
            "trajectory forces",
            "trajectory velocities",
            "trajectory system name",
            "make current",
            "trajectory extxyz",
            "trajectory extxyz skip frames",
            "trajectory extxyz filename",
            "trajectory extxyz append",
            "atomic positions rate",
            "atomic positions number of samples",
            "com positions rate",
            "com positions number of samples",
            "atomic velocities rate",
            "atomic velocities number of samples",
            "com velocities rate",
            "com velocities number of samples",
            "heat flux rate",
            "heat flux number of samples",
            "shear stress rate",
            "shear stress number of samples",
        ):
            if title == "Heat Flux" and (key == "heat flux" or "centroid" in key):
                continue
            self[key] = P[key].widget(tframe)

        self.reset_trajectory_frame()

        if title == "Heat Flux":
            return frame

        # Frame to isolate widgets
        c_frame = self["control_frame"] = ttk.LabelFrame(
            self["frame"],
            borderwidth=4,
            relief="sunken",
            text="General Parameters",
            labelanchor="n",
            padding=10,
        )

        for key in lammps_step.NVE_Parameters.parameters:
            if key == "control_properties":
                self[key] = P[key].widget(c_frame, metadata=self.property_metadata)
            else:
                self[key] = P[key].widget(c_frame)

        # make the control combobox wide enough
        self["run_control"].combobox.configure(width=40)

        # and binding to change as needed
        self["run_control"].combobox.bind(
            "<<ComboboxSelected>>", self.reset_control_frame
        )

        return frame

    def reset_dialog(self, widget=None):
        """Layout the widgets as needed for the current state"""

        frame = self["frame"]
        # Clear the dialog
        for slave in frame.grid_slaves():
            slave.grid_forget()

        row = 0
        # Put in our control frame
        self["control_frame"].grid(row=row, column=0)
        row += 1

        # and the widgets in it
        self.reset_control_frame()

        # And how to handle the structure
        if self.node.calculation == "nve":
            self["structure"].grid(row=row, column=0)
            row += 1

        return row

    def reset_control_frame(self, widget=None):
        """Layout the control widgets as needed for the current state"""

        run_control = self["run_control"].get()

        # Clear out the previous widgets
        c_frame = self["control_frame"]
        for slave in c_frame.grid_slaves():
            slave.grid_forget()

        # And put them back in depending...
        row = 0
        # self["run_control"].grid(row=row, column=0, sticky=tk.W)
        # row += 1

        widgets = []
        if "fixed length" in run_control:
            widgets.append(self["time"])
            self["time"].grid(row=row, column=0, sticky=tk.W)
            row += 1
        else:
            widgets.append(self["maximum_time"])
            self["maximum_time"].grid(row=row, column=0, sticky=tk.W)
            row += 1
        widgets.append(self["timestep"])
        self["timestep"].grid(row=row, column=0, sticky=tk.W)
        row += 1
        if "fixed length" in run_control:
            widgets.append(self["sampling"])
            self["sampling"].grid(row=row, column=0, sticky=tk.W)
            row += 1
        else:
            self["control_properties"].grid(row=row, column=0, sticky=tk.NSEW)
            row += 1

        sw.align_labels(widgets, sticky=tk.E)

    def reset_trajectory_frame(self, widget=None):
        """Layout the trajectory frame according to its contents."""
        trajectory = self["trajectory"].get()
        save = self["trajectory save"].get()
        export = self["trajectory export"].get()
        atomic_positions = self["atomic positions"].get()
        com_positions = self["com positions"].get()
        atomic_velocities = self["atomic velocities"].get()
        com_velocities = self["com velocities"].get()
        heat_flux = self["heat flux"].get()
        shear_stress = self["shear stress"].get()

        # Clear out the previous widgets
        frame = self["trajectory frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        row = 0
        widgets = []
        widgets1 = []
        widgets2 = []
        widgets3 = []
        widgets4 = []

        self["trajectory"].grid(row=row, column=0, columnspan=5, sticky=tk.EW)
        widgets.append(self["trajectory"])
        row += 1
        if trajectory != "never":
            if self.is_expr(trajectory) or "number" in trajectory:
                self["trajectory number of samples"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["trajectory number of samples"])
                row += 1
            if self.is_expr(trajectory) or "interval" in trajectory:
                self["trajectory rate"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["trajectory rate"])
                row += 1
            for key in (
                "trajectory forces",
                "trajectory velocities",
                "trajectory save",
            ):
                self[key].grid(row=row, column=1, columnspan=4, sticky=tk.EW)
                widgets1.append(self[key])
                row += 1

            if save != "no":
                for key in ("trajectory system name", "make current"):
                    self[key].grid(row=row, column=2, columnspan=3, sticky=tk.EW)
                    widgets2.append(self[key])
                    row += 1

            for key in ("trajectory export",):
                self[key].grid(row=row, column=1, columnspan=4, sticky=tk.EW)
                widgets1.append(self[key])
                row += 1

            if export != "no":
                for key in ("trajectory extxyz",):
                    self[key].grid(row=row, column=2, columnspan=3, sticky=tk.EW)
                    widgets2.append(self[key])
                    row += 1
                for key in (
                    "trajectory extxyz skip frames",
                    "trajectory extxyz filename",
                    "trajectory extxyz append",
                ):
                    self[key].grid(row=row, column=3, columnspan=2, sticky=tk.EW)
                    widgets3.append(self[key])
                    row += 1

        self["atomic positions"].grid(row=row, column=0, columnspan=5, sticky=tk.EW)
        widgets.append(self["atomic positions"])
        row += 1
        if atomic_positions != "never":
            if self.is_expr(atomic_positions) or "number" in atomic_positions:
                self["atomic positions number of samples"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["atomic positions number of samples"])
                row += 1
            if self.is_expr(atomic_positions) or "interval" in atomic_positions:
                self["atomic positions rate"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["atomic positions rate"])
                row += 1

        self["com positions"].grid(row=row, column=0, columnspan=5, sticky=tk.EW)
        widgets.append(self["com positions"])
        row += 1
        if com_positions != "never":
            if self.is_expr(com_positions) or "number" in com_positions:
                self["com positions number of samples"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["com positions number of samples"])
                row += 1
            if self.is_expr(com_positions) or "interval" in com_positions:
                self["com positions rate"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["com positions rate"])
                row += 1

        self["atomic velocities"].grid(row=row, column=0, columnspan=5, sticky=tk.EW)
        widgets.append(self["atomic velocities"])
        row += 1
        if atomic_velocities != "never":
            if self.is_expr(atomic_velocities) or "number" in atomic_velocities:
                self["atomic velocities number of samples"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["atomic velocities number of samples"])
                row += 1
            if self.is_expr(atomic_velocities) or "interval" in atomic_velocities:
                self["atomic velocities rate"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["atomic velocities rate"])
                row += 1

        self["com velocities"].grid(row=row, column=0, columnspan=5, sticky=tk.EW)
        widgets.append(self["com velocities"])
        row += 1
        if com_velocities != "never":
            if self.is_expr(com_velocities) or "number" in com_velocities:
                self["com velocities number of samples"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["com velocities number of samples"])
                row += 1
            if self.is_expr(com_velocities) or "interval" in com_velocities:
                self["com velocities rate"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["com velocities rate"])
                row += 1

        if self.__class__.__name__ == "TkHeatFlux":
            pass
        else:
            self["heat flux"].grid(row=row, column=0, columnspan=5, sticky=tk.EW)
            widgets.append(self["heat flux"])
            row += 1
            if heat_flux != "never":
                if self.is_expr(heat_flux) or "number" in heat_flux:
                    self["heat flux number of samples"].grid(
                        row=row, column=1, columnspan=4, sticky=tk.EW
                    )
                    widgets1.append(self["heat flux number of samples"])
                    row += 1
                if self.is_expr(heat_flux) or "interval" in heat_flux:
                    self["heat flux rate"].grid(
                        row=row, column=1, columnspan=4, sticky=tk.EW
                    )
                    widgets1.append(self["heat flux rate"])
                    row += 1

        self["shear stress"].grid(row=row, column=0, columnspan=5, sticky=tk.EW)
        widgets.append(self["shear stress"])
        row += 1
        if shear_stress != "never":
            if self.is_expr(shear_stress) or "number" in shear_stress:
                self["shear stress number of samples"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["shear stress number of samples"])
                row += 1
            if self.is_expr(shear_stress) or "interval" in shear_stress:
                self["shear stress rate"].grid(
                    row=row, column=1, columnspan=4, sticky=tk.EW
                )
                widgets1.append(self["shear stress rate"])
                row += 1

        width = sw.align_labels(widgets, sticky=tk.E)
        width1 = sw.align_labels(widgets1, sticky=tk.E)
        indent = 75
        size = max(width - width1 + indent, 0)
        frame.columnconfigure(0, minsize=size)
        if len(widgets2) > 0:
            width2 = sw.align_labels(widgets2, sticky=tk.E)
            size = max(width1 - width2 + indent, 0)
            frame.columnconfigure(1, minsize=size)
            if len(widgets3) > 0:
                width3 = sw.align_labels(widgets3, sticky=tk.E)
                size = max(width2 - width3 + indent, 0)
                frame.columnconfigure(2, minsize=size)
                if len(widgets4) > 0:
                    width4 = sw.align_labels(widgets4, sticky=tk.E)
                    size = max(width3 - width4 + indent, 0)
                    frame.columnconfigure(2, minsize=size)
        frame.columnconfigure(4, weight=1)

    def handle_dialog(self, result):
        if result == "OK":
            # Shortcut for parameters
            P = self.node.parameters

            value, units = self["time"].get()
            P["time"].value = value
            P["time"].units = units

            tmp = self["timestep"].get()
            if tmp in P["timestep"].enumeration:
                P["timestep"].value = tmp
            else:
                P["timestep"].value = tmp[0]
                P["timestep"].units = tmp[1]

            tmp = self["sampling"].get()
            if tmp in P["sampling"].enumeration:
                P["sampling"].value = tmp
            else:
                P["sampling"].value = tmp[0]
                P["sampling"].units = tmp[1]

        # Let base classes reap their parameters
        super().handle_dialog(result)



# -*- coding: utf-8 -*-
"""Control parameters for the initialization in LAMMPS"""

import logging
import seamm

logger = logging.getLogger(__name__)

kspace_methods = {
    "automatic": "",
    "none": "none",
    "Ewald summation method": "ewald {kspace_accuracy}",
    "PPPM (Particle-particle particle-mesh) method": "pppm {kspace_accuracy}",
    "PPPM method for few charged atoms": "pppm/cg {kspace_accuracy} {smallq}",
    "PPPM method with a staggered mesh": "pppm/stagger {kspace_accuracy}",
    "PPPM method including dispersion terms": "pppm/disp {kspace_accuracy}",
    "MSM (Multilevel summation method)": "msm {kspace_accuracy}",
    "MSM method for few charged atoms": "msm/cg {kspace_accuracy} {smallq}",
}

charge_methods = {
    "default for forcefield": "default",
    "no charges": "none",
    "charge equilibration (QEq), older reaxff implementation": "qeq/reaxff",
    "charge equilibration (QEq) with point charges": "qeq/point",
    "charge equilibration (QEq) with shielded charges": "qeq/shielded",
    "charge equilibration with atom-condensed Kohn–Sham DFT (ACKS2)": "acks2/reaxff",
    "charge transfer with polarization current equilibration (QTPIE)": "qtpie/reaxff",
}


class InitializationParameters(seamm.Parameters):
    """The control parameters for the initialization in LAMMPS"""

    parameters = {
        "cutoff": {
            "default": 10.0,
            "kind": "float",
            "default_units": "Å",
            "enumeration": tuple(),
            "format_string": ".1f",
            "description": "Cutoff:",
            "help_text": (
                "The cutoff used for non-bonded and similar " "interactions."
            ),
        },
        "kspace_method": {
            "default": "automatic",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": tuple(kspace_methods.keys()),
            "format_string": "",
            "description": "K-space method:",
            "help_text": "The method for handling long-range interactions.",
        },
        "kspace_accuracy": {
            "default": 1.0e-05,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": ".2e",
            "description": "K-space accuracy:",
            "help_text": "The target accuracy for the k-space method.",
        },
        "kspace_smallq": {
            "default": 1.0e-05,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": ".1e",
            "description": "K-space negligable charge:",
            "help_text": (
                "The cutoff for the charge on an atom to be considered not zero."
            ),
        },
        "atomic charges": {
            "default": "default for forcefield",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": tuple(charge_methods.keys()),
            "format_string": "",
            "description": "Atomic charge handling:",
            "help_text": "The method for handling atomic charges.",
        },
        "qeq convergence": {
            "default": 1.0e-06,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": ".2e",
            "description": "QEq convergence:",
            "help_text": "The covergence goal for the charge equilibration.",
        },
        "qeq iterations": {
            "default": 100,
            "kind": "integer",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "d",
            "description": "QEq iterations:",
            "help_text": (
                "The maximun number of iterations for the charge equilibration."
            ),
        },
        "charged_atom_fraction_cutoff": {
            "default": 0.1,
            "kind": "float",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": ".2f",
            "description": "Cutoff for charged atoms:",
            "help_text": (
                "The upper limit of the fraction of charged atoms "
                "for using methods that exploit sparsity of charges."
            ),
        },
        "ewald_atom_cutoff": {
            "default": 1000,
            "kind": "integer",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "d",
            "description": "Atom limit for using Ewald summations:",
            "help_text": (
                "The upper limit of the number of atoms for using "
                "Ewald summations by default."
            ),
        },
        "msm_atom_cutoff": {
            "default": 5000,
            "kind": "integer",
            "default_units": "",
            "enumeration": tuple(),
            "format_string": "d",
            "description": "Lower atom limit for using the MSM method:",
            "help_text": (
                "The lower limit of the number of atoms for using "
                "the MSM method by default."
            ),
        },
        "tail_correction": {
            "default": "yes",
            "kind": "boolean",
            "format_string": "s",
            "enumeration": ("no", "yes"),
            "description": "Add tail corrections",
            "help_text": (
                "Add tail corrections for neglected long-range " "van der Waals terms."
            ),
        },
        "shift_nonbond": {
            "default": "no",
            "kind": "boolean",
            "format_string": "s",
            "enumeration": ("no", "yes"),
            "description": ("Shift nonbonded interactions to zero at the " "cutoff"),
            "help_text": (
                "Shift the nonbonded interactions to zero at the "
                "cutoff distance to avoid energy jumps."
            ),
        },
        "rigid_waters": {
            "default": "yes",
            "kind": "boolean",
            "format_string": "s",
            "enumeration": ("no", "yes"),
            "description": ("Keep the internal geometry of water molecules " "fixed"),
            "help_text": (
                "Use shake or rattle, as appropriate, to keep "
                "the water molecules rigid."
            ),
        },
        "fix_XH_bond_lengths": {
            "default": "none",
            "kind": "enumeration",
            "default_units": "",
            "enumeration": ("none", "all", "CH"),
            "format_string": "",
            "description": "Fix X-H bond lengths",
            "help_text": "Fix the bond lengths for H bonded to X.",
        },
    }

    def __init__(self, defaults={}, data=None):
        """Initialize the instance, by default from the default
        parameters given in the class"""

        super().__init__(
            defaults={**InitializationParameters.parameters, **defaults}, data=data
        )



# -*- coding: utf-8 -*-

"""NVE (microcanonical) dynamics in LAMMPS"""

import bz2
import gzip
import json
import os
from pathlib import Path
import traceback

import numpy as np
from tabulate import tabulate

import lammps_step
from lammps_step import from_lammps_units
import logging
import seamm
from seamm_util import ureg, Q_, units_class  # noqa: F401
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("lammps")


class NVE(lammps_step.Energy):
    def __init__(
        self,
        flowchart=None,
        title="NVE dynamics",
        extension=None,
        logger=logger,
    ):
        """Initialize the node"""

        logger.debug("Creating NVE {}".format(self))

        super().__init__(
            flowchart=flowchart,
            title=title,
            extension=extension,
            logger=logger,
        )
        self.logger.debug("NVE.init() creating NVE_Parameters object")

        self._calculation = "nve"
        self._metadata = lammps_step.metadata
        self.parameters = lammps_step.NVE_Parameters()

    def analyze(self, data={}, properties=None, table=None, output=""):
        """Analyze the results of the simulation.

        Parameters
        ----------
        """
        # Calculate the enthalpy of formation, if possible
        tmp_text = self.calculate_enthalpy_of_formation(data)
        if tmp_text != "":
            path = self.wd / "Thermochemistry.txt"
            path.write_text(tmp_text)

        # Print the results from the analysis of the state trajectory
        for i, _property in enumerate(table["Property"]):
            if _property.startswith("Epe"):
                stderr = table["StdErr"][i]
                i += 1
                new = {k: v[0:i] for k, v in table.items()}
                for key in ("DfE0", "E atomization"):
                    if key in data:
                        new["Property"].append(key)
                        new["Value"].append(f"{data[key]:.2f}")
                        new[" "].append("±")
                        new["StdErr"].append(stderr)
                        new["Units"].append("kcal/mol")
                        new["convergence"].append("")
                        new["tau"].append("")
                        new["inefficiency"].append("")
                for key, value in table.items():
                    new[key].extend(table[key][i:])
                table = new
                break
        else:
            for key in ("DfE0", "E atomization"):
                if key in data:
                    table["Property"].append(key)
                    table["Value"].append(f"{data[key]:.2f}")
                    table[" "].append("±")
                    table["StdErr"].append("")
                    table["Units"].append("kcal/mol")
                    table["convergence"].append("")
                    table["tau"].append("")
                    table["inefficiency"].append("")

        # Print out a table of results.
        text = ""
        tmp = tabulate(
            table,
            headers="keys",
            tablefmt="simple",
            disable_numparse=True,
            colalign=(
                "center",
                "decimal",
                "center",
                "decimal",
                "left",
                "decimal",
                "decimal",
                "decimal",
            ),
        )
        length = len(tmp.splitlines()[0])
        text += "\n"
        text += "Properties".center(length)
        text += "\n"
        text += tmp
        text += "\n"

        printer.normal(__(text, indent=8 * " ", wrap=False, dedent=False))

        have_warning = False
        for value in table["Property"]:
            if len(value) > 0 and value[-1] == "*":
                have_warning = True
                break

        have_acf_warning = False
        for value in table["tau"]:
            if len(value) > 0 and value[-1] == "^":
                have_acf_warning = True
                break

        if have_warning:
            printer.normal(
                __(
                    "          * this property has less than 100 independent "
                    "samples, so may not be accurate.",
                    wrap=False,
                    dedent=False,
                )
            )

        if have_acf_warning:
            printer.normal(
                __(
                    "          ^ there are not enough samples after "
                    "equilibration to plot the ACF.",
                    wrap=False,
                    dedent=False,
                )
            )

        if have_warning or have_acf_warning:
            printer.normal("\n")

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )

        # Get the initial structure
        _, initial_configuration = self.get_system_configuration()

        # Handle the new structure as needed
        system, configuration = self.get_system_configuration(P, model=self.model)

        # Read the dump file and get the structure
        dump_file = Path(self.directory) / f"{self.calculation}.dump"
        if dump_file.exists:
            try:
                xyz, fractional, cell, vxyz = self.parent.read_dump(dump_file)
            except Exception as e:
                xyz = None
                cell = None
                vxyz = None
                printer.normal("Warning: unable to read the LAMMPS dumpfile")
                logger.warning(f"The was a problem reading the LAMMPS dumpfile: {e}")
                logger.warning(traceback.format_exc())
        else:
            printer.normal("Warning: there is no 'dump' file from LAMMPS")
            xyz = None
            cell = None
            vxyz = None

        if configuration is not None:
            if cell is not None:
                configuration.cell.parameters = cell
            if xyz is not None:
                configuration.atoms.set_coordinates(xyz, fractionals=fractional)
            if vxyz is not None:
                # LAMMPS only has Cartesian velocities
                configuration.atoms.set_velocities(vxyz, fractionals=False)

            # And the name of the configuration.
            text = seamm.standard_parameters.set_names(
                system,
                configuration,
                P,
                _first=True,
                model=self.model,
            )
            printer.normal(__(text, **data, indent=self.indent + 4 * " "))
            printer.normal("")

        # Save the trajectory to a new system and its configurations
        if P["trajectory save"]:
            text = self._save_trajectory(P)
            printer.normal(__(text, indent=self.indent + 4 * " "))
            printer.normal("")

        # Import or convert the trajectory.
        if P["trajectory export"]:
            text = self._export_trajectory(P)
            printer.normal(__(text, indent=self.indent + 4 * " "))
            printer.normal("")

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """

        if not P:
            P = self.parameters.values_to_dict()

        model = self.model

        text = "{time} of microcanonical (NVE) dynamics using a "
        text += "timestep of {timestep}"
        if model is None:
            text += "."
        else:
            text += f" using the {model} forcefield. "
        text += "The trajectory will be sampled every {sampling}."

        return self.header + "\n" + __(text, **P, indent=4 * " ").__str__()

    def _export_trajectory(self, P):
        """Export the trajectory in requested formats.

        Parameters
        ----------
        P : dict(str, value)
            The control parameters for this step

        Returns
        -------
        str
            A text string for printing, containing warnings, etc.
        """
        text = ""
        dump = Path(self.directory) / "trajectory.dump_trj.gz"
        trj = Path(self.directory) / "trajectory.trj"

        if not dump.exists() and not trj.exists():
            return (
                "Neither the coordinate or energy part of the trajectory exists, so "
                "cannot export it."
            )

        if P["trajectory extxyz"]:
            text += self._convert_trajectory_to_extxyz(dump, trj, P)

        return text

    def _convert_trajectory_to_extxyz(self, dump, trj, P):
        """Convert the trajectory to the ASE extended XYZ format.

        Parameters
        ----------
        dump : Path
            The path to the dump part of the trajectory.
        trj : Path
            The path to the state (energy, pressure) part of the trajectory
        P : dict(str, value)
            The control parameters for this step

        Returns
        -------
        str
            Text for printing explaining the process
        """
        text = ""
        if not dump.exists():
            return (
                "The coordinate dump of the trajectory does not exist, so cannot "
                "create the extxyz file."
            )
        if trj.exists():
            trj_data = trj.read_text().splitlines()
            header = trj_data[0]
            if not header.startswith("!MolSSI trajectory"):
                trj_data = None
                text = (
                    "The file for the energies etc. for the trajectory exists "
                    "but is not the right type of file, so the energies, etc. "
                    "will not be in the extxyz file"
                )
            else:
                state_variables = trj_data[1].split()
                trj_data = [i.split() for i in trj_data[2:]]
        else:
            trj_data = None
            text = (
                "The energies etc. for the trajectory are missing so they will "
                "not be in the extxyz file"
            )

        # Do it!
        _, timestep = self.timestep(P["timestep"])
        _, configuration = self.get_system_configuration()
        symbols = configuration.atoms.symbols

        filename = P["trajectory extxyz filename"].strip()
        if filename.startswith("/"):
            path = Path(self.flowchart.root_directory) / filename[1:]
        else:
            path = self.wd / Path(filename)

        nskip = P["trajectory extxyz skip frames"]
        frame = 0
        start_line = 0
        lines = []

        append = P["trajectory extxyz append"]
        mode = "a" if append else "w"
        text += "Appended" if append else "Wrote"
        with (
            gzip.open(dump, "rt") as fdin,
            (
                gzip.open(path, mode=mode + "t")
                if path.suffix == ".gz"
                else (
                    bz2.open(path, mode=mode + "t")
                    if path.suffix == ".bz2"
                    else open(path, mode)
                )
            ) as fdout,
        ):
            for line in fdin:
                line = line.strip()
                if line.startswith("ITEM: TIMESTEP"):
                    if len(lines) > 0:
                        frame += 1
                        if frame > nskip:
                            results = self._parse_dump_frame(lines, start_line)
                            results["symbols"] = symbols
                            results.update(
                                {
                                    k: v
                                    for k, v in zip(
                                        state_variables, trj_data[frame - 1]
                                    )
                                }
                            )
                            extxyz = self._to_extxyz(results, timestep=timestep)
                            fdout.write(extxyz)
                        start_line += len(lines)
                        lines = []
                lines.append(line)
            # Process last frame
            if len(lines) > 0:
                frame += 1
                if frame > nskip:
                    results = self._parse_dump_frame(lines, start_line)
                    results["symbols"] = symbols
                    results.update(
                        {k: v for k, v in zip(state_variables, trj_data[frame - 1])}
                    )
                    extxyz = self._to_extxyz(results, timestep=timestep)
                    fdout.write(extxyz)
        if nskip > 0:
            text += (
                f" {frame - nskip} frames of the trajectory to {path}, "
                f"skipping the first {nskip} frames."
            )
        else:
            text += f" {frame - nskip} frames of the trajectory to {path}."
        return text

    def get_input(self, extras=None):
        """Get the input for an NVE dynamics run in LAMMPS"""

        # See what type of forcefield we have and handle it
        ff_form = self.parent.ff_form()

        self.description = []
        self.description.append(__(self.header, indent=4 * " "))

        P = self.parameters.current_values_to_dict(
            context=seamm.flowchart_variables._data
        )
        _, configuration = self.get_system_configuration()

        timestep, P["timestep"] = self.timestep(P["timestep"])

        if extras is not None and "nsteps" in extras:
            nsteps = extras["nsteps"]
        else:
            time = P["time"].to("fs").magnitude
            nsteps = round(time / timestep)

        # Have to fix formatting for printing...
        PP = dict(P)
        for key in PP:
            if isinstance(PP[key], units_class):
                PP[key] = "{:~P}".format(PP[key])

        self.description.append(__(self.description_text(), **PP, indent=7 * " "))

        # time = lammps_step.to_lammps_units(P['time'], quantity='time')
        # nsteps = round(time / timestep)

        thermo_properties = (
            "time temp press etotal ke pe ebond "
            "eangle edihed eimp evdwl etail ecoul elong"
        )
        properties = "v_time v_temp v_press v_etotal v_ke v_pe v_emol v_epair"
        title2 = "tstep t T P Etot Eke Epe Emol Epair"
        if configuration.periodicity == 3:
            properties += " v_sxx v_syy v_szz v_syz v_sxz v_sxy"
            title2 += " Sxx Syy Szz Syz Sxz Sxy"
        if self.parent.have_dreiding_hbonds:
            thermo_properties += " v_N_hbond v_E_hbond"
            properties += " v_N_hbond v_E_hbond"
            title2 += " N_hbond E_hbond"

        lines = []
        nfixes = 0
        ncomputes = 0
        ndumps = 0
        lines.append("")
        lines.append(f"# {self.header}")
        lines.append("")
        lines.append("reset_timestep      0")
        lines.append("timestep            {}".format(timestep))
        lines.append("thermo_style        custom {}".format(thermo_properties))
        lines.append("thermo              {}".format(int(nsteps / 100)))
        nfixes += 1
        lines.append("fix                 {} all nve".format(nfixes))

        # For the heat flux, if requested, we need extra input
        if P["heat flux"] != "never":
            # Unit conversion factor
            if lammps_step.get_lammps_unit_system() == "metal":
                factor = Q_("eV/Å^2/ps")
            else:
                factor = (
                    Q_("kcal/Å^2/fs/mol") / Q_("kcal/mol") * Q_("kcal/mol").to("kJ")
                )
            factor = factor.m_as("W/m^2")
            if ff_form == "class2" or not P["use centroid stress"]:
                # Centroid/stress/atom does not handle class2 ff ... cross-terms?
                lines.append(
                    f"""
compute             KE all ke/atom
compute             PE all pe/atom

compute             S_p all stress/atom NULL virial
compute             flux_p all heat/flux KE PE S_p

#          Conversion from kcal/Å^2/fs/mol to W/m^2

variable            factor equal {factor}
variable            Jx equal v_factor*c_flux_p[1]/vol
variable            Jy equal v_factor*c_flux_p[2]/vol
variable            Jz equal v_factor*c_flux_p[3]/vol
"""
                )
            else:
                lines.append(
                    f"""
compute             KE all ke/atom
compute             PE all pe/atom

#          centroid doesn't work with kspace, so split into pair and non-pair parts

compute             S_p all stress/atom NULL pair kspace
compute             S_b all centroid/stress/atom NULL bond angle dihedral improper
compute             flux_p all heat/flux KE PE S_p
compute             flux_b all heat/flux KE PE S_b

#          Conversion from kcal/Å^2/fs/mol to W/m^2

variable            factor equal {factor}
variable            Jx equal v_factor*(c_flux_p[1]+c_flux_b[1])/vol
variable            Jy equal v_factor*(c_flux_p[2]+c_flux_b[2])/vol
variable            Jz equal v_factor*(c_flux_p[3]+c_flux_b[3])/vol
"""
                )

        # summary output written 10 times during run so we can see progress
        nevery = 10
        nfreq = int(nsteps / 10)
        nrepeat = int(nfreq / nevery)
        nfreq = nevery * nrepeat
        nfixes += 1
        filename = f"@{self._id[-1]}+nve_summary.trj"
        lines.append(
            f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
            f"                       {properties} &\n"
            "                       off 2 &\n"
            f"                       file {filename}"
        )

        # instantaneous output written for averaging
        if P["sampling"] == "none":
            self.description.append(
                __(
                    "The run will be {nsteps:n} steps of dynamics.",
                    nsteps=nsteps,
                    indent=7 * " ",
                )
            )
        else:
            sampling = lammps_step.to_lammps_units(P["sampling"], quantity="time")
            nevery = max(1, round(sampling / timestep))
            nfreq = int(nsteps / nevery)
            nrepeat = 1
            nfreq = nevery * nrepeat
            nfixes += 1
            dt = (nevery * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "NVE",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // nevery,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI trajectory 2.0 " + text
            filename = f"@{self._id[-1]}+nve_state.trj"
            lines.append(
                f"fix                 {nfixes} all ave/time {nevery} 1 {nfreq} &\n"
                f"                       {properties} &\n"
                "                       off 2 &\n"
                f"                       title1 '{title1}' &\n"
                f"                       title2 '{title2}' &\n"
                f"                       file {filename}"
            )
            self.description.append(
                __(
                    (
                        "The run will be {nsteps:,d} steps of dynamics "
                        "sampled every {nevery:n} steps."
                    ),
                    nsteps=nsteps,
                    nevery=nevery,
                    indent=7 * " ",
                )
            )

        # Handle trajectories
        tmp, ncomputes, ndumps, nfixes = self.trajectory_input(
            P, timestep, nsteps, ncomputes, ndumps, nfixes
        )
        lines.extend(tmp)

        if extras is not None and "shake" in extras:
            nfixes += 1
            lines.append(extras["shake"].format(nfixes))

        lines.append("")
        lines.append("run                 {}".format(nsteps))
        lines.append("")
        filename = f"@{self._id[-1]}+nve.dump"
        if configuration.periodicity == 0:
            lines.append(
                f"write_dump         all custom  {filename} id xu yu zu vx vy vz"
                " modify flush yes sort id"
            )
        else:
            lines.append(
                f"write_dump         all custom  {filename} id xsu ysu zsu vx vy vz"
                " modify flush yes sort id"
            )
        lines.append("")

        for i in range(1, ncomputes + 1):
            lines.append(f"uncompute           {i}")
        for i in range(1, ndumps + 1):
            lines.append(f"undump              {i}")
        for i in range(1, nfixes + 1):
            lines.append(f"unfix               {i}")
        if P["heat flux"] != "never":
            if ff_form != "class2" and P["use centroid stress"]:
                lines.append("uncompute           flux_b")
                lines.append("uncompute           S_b")
            lines.append("uncompute           flux_p")
            lines.append("uncompute           S_p")
            lines.append("uncompute           PE")
            lines.append("uncompute           KE")
            lines.append("variable            factor delete")
            lines.append("variable            Jx delete")
            lines.append("variable            Jy delete")
            lines.append("variable            Jz delete")
        lines.append("")

        return {
            "script": lines,
            "postscript": None,
            "use python": False,
        }

    def timestep(self, value):
        """Get the timestep in the correct units.

        This handles the 'normal', 'accurate' and 'coarse' values,
        which depend on the mass in an empirical fashion.

        Parameters
        ----------
        value : str or Pint quantity
            The desired timestep, which may be a Pint quantity with units
            or one of 'notmal', 'accurate but slow', or 'coarse but fast'

        Returns
        -------
        timestep : float
            The magnitude of the time step in the appropriate LAMMPS units
        """
        masses = self.parent._data["masses"]
        min_mass = min(masses)

        if self.parent.ff_form() == "reaxff":
            # ReaxFF needs a smaller timestep for the QEq part
            factor = 0.5
        else:
            # These are based on masses as a proxy for vibrational frequencies
            if min_mass < 10:
                factor = 1
            elif min_mass < 50:
                factor = 2
            else:
                factor = 4

        if value == "normal":
            timestep = 1.0 * factor
            value = Q_(timestep, ureg.fs)
        elif value == "accurate but slow":
            timestep = 0.5 * factor
            value = Q_(timestep, ureg.fs)
        elif value == "coarse but fast":
            timestep = 2.0 * factor
            value = Q_(timestep, ureg.fs)
        else:
            value = Q_(value)

        timestep = lammps_step.to_lammps_units(value, quantity="time")

        return (timestep, value)

    def trajectory_input(self, P, timestep, nsteps, ncomputes, ndumps, nfixes):
        """Create the part of the input handling the trajectories.

        Parameters
        ----------
        P : dict
            The dictionary of options
        timestep : int
            The timestep in LAMMPS units
        nsteps : int
            Total number of steps in the run
        ncomputes : int
            The counter for the computes
        ndumps : int
            The counter for the dumps
        nfixes : int
            The counter for the fixes

        Returns
        -------
        [str]
            The input lines as array
        int
            The counter for the computes
        int
            The counter for the dumps
        int
            The counter for the fixes
        """
        lines = []
        sampling = P["trajectory"]
        if sampling != "never":
            if "interval" in sampling:
                t_s = lammps_step.to_lammps_units(P["trajectory rate"], quantity="time")
                n = max(1, round(t_s / timestep))
            else:
                n = max(1, nsteps // P["trajectory number of samples"])
            ndumps += 1
            filename = f"@{self._id[-1]}+trajectory.dump_trj.gz"
            if "charges" in self.parent.eex:
                line = (
                    f"dump                {ndumps} all custom {n} {filename} id element"
                    " q xu yu zu"
                )
            else:
                line = (
                    f"dump                {ndumps} all custom {n} {filename} id element"
                    " xu yu zu"
                )
            if P["trajectory forces"]:
                line += " fx fy fz"
            if P["trajectory velocities"]:
                line += " vx vy vz"
            lines.append("\n")
            lines.append(line)
            tmp = " ".join(self.parent.eex["elements"])
            lines.append(
                f"dump_modify         {ndumps} sort id format float %.5f element {tmp}"
            )

            if False:
                # Temporarily put in dump to extxyz
                ndumps += 1
                filename = f"@{self._id[-1]}+trajectory.extxyz.gz"
                line = f"dump                {ndumps} all extxyz {n} {filename}"
                lines.append(line)
                element_string = " ".join(self.parent.eex["elements"])
                lines.append(f"dump_modify         {ndumps} element {element_string}")

            # The energies, stress, etc
            properties = "v_time v_temp v_press v_etotal v_ke v_pe v_emol v_epair"
            title2 = "tstep t T P Etot Eke Epe Emol Epair"
            if self.parent.have_dreiding_hbonds:
                properties += " v_N_hbond v_E_hbond"
                title2 += " N_hbond E_hbond"

            filename = f"@{self._id[-1]}+trajectory.trj"
            nfixes += 1
            dt = (n * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "state for trajectory",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // n,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI trajectory 2.0 " + text
            title2 += " Sxx Syy Szz Syz Sxz Sxy"
            properties += " v_sxx v_syy v_szz v_syz v_sxz v_sxy"
            lines.append(
                "\n"
                f"fix                 {nfixes} all ave/time {n} 1 {n} "
                f"                        {properties}    &\n"
                f"                        title1 '{title1}' &\n"
                f"                        title2 '{title2}' &\n"
                f"                        file {filename}"
            )
        sampling = P["atomic positions"]
        if sampling != "never":
            if "interval" in sampling:
                t_s = lammps_step.to_lammps_units(
                    P["atomic positions rate"], quantity="time"
                )
                n = max(1, round(t_s / timestep))
            else:
                n = max(1, nsteps // P["atomic positions number of samples"])
            ndumps += 1
            filename = f"@{self._id[-1]}+atomic_positions.dump_trj"
            lines.append(
                "\n"
                f"dump                {ndumps} all custom {n} {filename} id xu yu zu\n"
                f"dump_modify         {ndumps} sort id"
            )
        sampling = P["com positions"]
        if sampling != "never":
            if "interval" in sampling:
                t_s = lammps_step.to_lammps_units(
                    P["com positions rate"], quantity="time"
                )
                n = max(1, round(t_s / timestep))
            else:
                n = max(1, nsteps // P["com positions number of samples"])

            filename = f"@{self._id[-1]}+com_positions.trj"
            ncomputes += 1
            c1 = ncomputes
            ncomputes += 1
            c2 = ncomputes
            nfixes += 1
            dt = (n * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "com positions",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // n,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI vector_trajectory 2.0 " + text
            title2 = "! timestep n_molecules"
            title3 = "molecule com_x com_y com_z"
            lines.append(
                "\n"
                f"compute             {c1} all chunk/atom molecule\n"
                f"compute             {c2} all com/chunk {c1}\n"
                f"fix                 {nfixes} all ave/time {n} 1 {n} c_{c2}[*] &\n"
                f"                        title1 '{title1}' &\n"
                f"                        title2 '{title2}' &\n"
                f"                        title3 '{title3}' &\n"
                f"                        file {filename} mode vector"
            )
        sampling = P["atomic velocities"]
        if sampling != "never":
            if "interval" in sampling:
                t_s = lammps_step.to_lammps_units(
                    P["atomic velocities rate"], quantity="time"
                )
                n = max(1, round(t_s / timestep))
            else:
                n = max(1, nsteps // P["atomic velocities number of samples"])
            ndumps += 1
            filename = f"@{self._id[-1]}+atomic_velocities.dump_trj"
            lines.append(
                "\n"
                f"dump                {ndumps} all custom {n} {filename} id vx vy vz\n"
                f"dump_modify         {ndumps} sort id"
            )
        sampling = P["com velocities"]
        if sampling != "never":
            if "interval" in sampling:
                t_s = lammps_step.to_lammps_units(
                    P["com velocities rate"], quantity="time"
                )
                n = max(1, round(t_s / timestep))
            else:
                n = max(1, nsteps // P["com velocities number of samples"])
            filename = f"@{self._id[-1]}+com_velocities.trj"
            ncomputes += 1
            c1 = ncomputes
            ncomputes += 1
            c2 = ncomputes
            nfixes += 1
            dt = (n * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "com velocities",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // n,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI vector_trajectory 2.0 " + text
            title2 = "! timestep n_molecules"
            title3 = "molecule com_vx com_vy com_vz"
            lines.append(
                "\n"
                f"compute             {c1} all chunk/atom molecule\n"
                f"compute             {c2} all vcm/chunk {c1}\n"
                f"fix                 {nfixes} all ave/time {n} 1 {n} c_{c2}[*] &\n"
                f"                        title1 '{title1}' &\n"
                f"                        title2 '{title2}' &\n"
                f"                        title3 '{title3}' &\n"
                f"                        file {filename} mode vector"
            )
        sampling = P["heat flux"]
        if sampling != "never":
            if "interval" in sampling:
                t_s = lammps_step.to_lammps_units(P["heat flux rate"], quantity="time")
                n = max(1, round(t_s / timestep))
            else:
                n = max(1, nsteps // P["heat flux number of samples"])
            filename = f"@{self._id[-1]}+heat_flux.trj"
            nfixes += 1
            dt = (n * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "heat flux",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // n,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI trajectory 2.0 " + text
            title2 = "Jx Jy Jz"
            lines.append(
                "\n"
                f"fix                 {nfixes} all ave/time {n} 1 {n} "
                "v_Jx v_Jy v_Jz &\n"
                f"                        title1 '{title1}' &\n"
                f"                        title2 '{title2}' &\n"
                f"                        file {filename}"
            )
        sampling = P["shear stress"]
        if sampling != "never":
            if "interval" in sampling:
                t_s = lammps_step.to_lammps_units(
                    P["shear stress rate"], quantity="time"
                )
                n = max(1, round(t_s / timestep))
            else:
                n = max(1, nsteps // P["shear stress number of samples"])
            filename = f"@{self._id[-1]}+shear_stress.trj"
            nfixes += 1
            dt = (n * P["timestep"]).to_compact()
            text = json.dumps(
                {
                    "code": "LAMMPS",
                    "type": "shear stress",
                    "dt": dt.magnitude,
                    "tunits": str(dt.u),
                    "nsteps": nsteps // n,
                },
                separators=(",", ":"),
            )
            title1 = "!MolSSI trajectory 2.0 " + text
            title2 = "Pxy Pxz Pyz"
            lines.append(
                "\n"
                f"fix                 {nfixes} all ave/time {n} 1 {n} "
                "v_pxy v_pxz v_pyz &\n"
                f"                        title1 '{title1}' &\n"
                f"                        title2 '{title2}' &\n"
                f"                        file {filename}"
            )

        return lines, ncomputes, ndumps, nfixes

    def _save_trajectory(self, P):
        """Save the trajectory to configurations.

        Parameters
        ----------
        P : dict(str, value)
            The control parameters for this step

        Returns
        -------
        str
            A text string for printing, containing warnings, etc.
        """
        text = ""
        dump = Path(self.directory) / "trajectory.dump_trj.gz"
        trj = Path(self.directory) / "trajectory.trj"

        if not dump.exists() and not trj.exists():
            return (
                "Neither the coordinate or energy part of the trajectory exists, so "
                "cannot save it."
            )
        if trj.exists():
            trj_data = trj.read_text().splitlines()
            header = trj_data[0]
            if not header.startswith("!MolSSI trajectory"):
                trj_data = None
                text = (
                    "The file for the energies etc. for the trajectory exists "
                    "but is not the right type of file, so the energies, etc. "
                    "will not be saved with the configuration. "
                )
            else:
                state_variables = trj_data[1].split()
                trj_data = [i.split() for i in trj_data[2:]]
        else:
            trj_data = None
            text = (
                "The energies etc. for the trajectory are missing so they will "
                "not be saved with the configuration. "
            )

        # Get the elements from the starting configuration
        initial_system, initial_configuration = self.get_system_configuration()
        symbols = initial_configuration.atoms.symbols

        # Make a new system
        name = P["trajectory system name"]
        system_db = initial_system.system_db
        if name == "current":
            system = system_db.system
        else:
            if system_db.system_exists(name):
                system = system_db.get_system(name)
            else:
                system = system_db.create_system(name=name)
        if P["make current"]:
            system_db.system = system

        # Read the trajectory dump file and process
        _, timestep = self.timestep(P["timestep"])
        frame = 0
        start_line = 0
        lines = []

        with gzip.open(dump, "rt") as fdin:
            for line in fdin:
                line = line.strip()
                if line.startswith("ITEM: TIMESTEP"):
                    if len(lines) > 0:
                        frame += 1
                        results = self._parse_dump_frame(lines, start_line)
                        results["symbols"] = symbols
                        results.update(
                            {k: v for k, v in zip(state_variables, trj_data[frame - 1])}
                        )
                        # Make a new configuration
                        time = timestep * results["timestep"]
                        name = f"{time:.1f~P}"
                        configuration = system.copy_configuration(
                            configuration=initial_configuration, name=name
                        )
                        self._to_configuration(configuration, results)
                    start_line += len(lines)
                    lines = []
                lines.append(line)
            # Process last frame
            if len(lines) > 0:
                frame += 1
                results = self._parse_dump_frame(lines, start_line)
                results["symbols"] = symbols
                results.update(
                    {k: v for k, v in zip(state_variables, trj_data[frame - 1])}
                )
                # Make a new configuration
                time = timestep * results["timestep"]
                name = f"{time:.1f~P}"
                configuration = system.copy_configuration(
                    configuration=initial_configuration, name=name
                )
                self._to_configuration(configuration, results)
        text += f"Created {frame} configurations of system '{system.name}'"
        if P["make current"]:
            text += ", which was made the current system,"
        text += " from the trajectory."
        return text

    def _check_property(self, properties, _property, **kwargs):
        """Check if a property exists after substitution and create if necessary.

        Parameters
        ----------
        properties : molsystem._Properties
            The properties object

        _property : str
            The name of the property

        **kwargs : {str: str}
            Optional variables for substituting in the property name.
        """
        expanded_property = _property.format_map(kwargs)
        if not properties.exists(expanded_property):
            # Get the general property's info to create the model property.
            _type, units, description = properties.metadata(_property)
            properties.add(
                expanded_property,
                _type=_type,
                units=units,
                description=description.format_map(kwargs),
            )
        return expanded_property

    def _to_configuration(self, configuration, data):
        """Update a configuration with the data from a step in a trajectory.

        Parameters
        ----------
        configuration : molsystem._Configuration
            The configuration to update
        data : dict(str, any)
            Dictionary containing coordinates and other atom properties
        """
        if configuration.periodicity == 3:
            # Set the cell
            if "cell" in data:
                configuration.cell.parameters = data["cell"]

            # Stress
            if "Sxx" in data:
                sfact = from_lammps_units(1, "atm").magnitude
                stress = [
                    sfact * float(data["Sxx"]),
                    sfact * float(data["Syy"]),
                    sfact * float(data["Szz"]),
                    sfact * float(data["Syz"]),
                    sfact * float(data["Sxz"]),
                    sfact * float(data["Sxy"]),
                ]
                _property = self._check_property(
                    configuration.properties, "stress#LAMMPS#{model}", model=self.model
                )
                configuration.properties.put(_property, stress)

        # Coordinates
        if "abc" in data:
            configuration.atoms.set_coordinates(data["abc"], fractionals=True)
        elif "xyz" in data:
            configuration.atoms.set_coordinates(data["xyz"], fractionals=False)

        if "velocities" in data:
            # LAMMPS only has Cartesian velocities. Already in Å/fs
            tmp = np.array(data["velocities"])
            configuration.atoms.set_velocities(tmp, fractionals=False)

        if "gradients" in data:
            # LAMMPS only has Cartesian forces, already in kJ/mol/Å
            tmp = np.array(data["gradients"])
            configuration.atoms.set_gradients(tmp, fractionals=False)

        # Energy
        if "Epe" in data:
            _property = self._check_property(
                configuration.properties,
                "potential energy#LAMMPS#{model}",
                model=self.model,
            )
            tmp = from_lammps_units(float(data["Epe"]), "kcal/mol").magnitude
            configuration.properties.put(_property, tmp)

            # Calculate the energy of formation if we can
            tmp_data = {"Epe": tmp}
            self.calculate_enthalpy_of_formation(tmp_data)
            if "DfE0" in tmp_data:
                _property = self._check_property(
                    configuration.properties,
                    "DfE0#LAMMPS#{model}",
                    model=self.model,
                )
                configuration.properties.put(_property, tmp_data["DfE0"])

    def _to_extxyz(self, data, timestep=None):
        """Create the text of ASE extxyz format for a structure

        Parameters
        ----------
        data : dict(str, any)
            Dictionary containing symbols, coordinates, and other atom properties

        timestep : pint.units
            The timestep.

        Note
        ----
        Example:

        8
        Properties=species:S:1:pos:R:3:REF_forces:R:3 REF_energy=-9289.028102188924 pbc="F F F"
        C       -0.97142774       0.56951904       1.10325229       2.77158783      -3.19505603       0.10221193
        O       -0.27179274       1.07438707      -0.03295067      -1.40444169       0.58287428       2.02393798
        C        0.44642827       0.04566107      -0.56782067      -2.59873178       1.73475706       0.42253222
        O        0.22332226      -1.04711390       0.04433633      -1.29638003      -2.74941166       4.75917280
        C       -0.54656476      -0.74672896       1.23491836      -0.44334119       2.10839425      -2.18446253
        O        0.98560923       0.09795906      -1.57642770       3.16925789       1.02455543      -4.39760192
        H       -0.66579676      -1.59970891       1.88536036      -0.27973052       0.31498569       0.12677338
        H       -1.44834375       1.17851210       1.87373829       0.08523677       0.17712004      -0.85144738
        """  # noqa: E501
        lines = []
        lines.append(f"{data['n_atoms']}")

        # The lattice/properties line
        if "lattice" in data:
            # already in Å
            line = 'Lattice="'
            line += " ".join([f"{v:.5f}" for v in data["lattice"]])
            line += '"'
        line += " Properties=species:S:1:pos:R:3"
        if "gradients" in data:
            # currently in kJ/mol/Å
            line += ":REF_forces:R:3"
            ffact = -Q_(1, "kJ/mol/Å").m_as("eV/Å")
        if "velocities" in data:
            # currentl in Å/fs
            line += ":vel:R:3"
            vfact = Q_(1, "Å/fs").m_as("eV^0.5/amu^0.5")

        # Add in other properties, like the energy ... currently just energy and stress
        if "Epe" in data:
            # Calculate the energy of formation if we can
            tmp_data = {
                "Epe": from_lammps_units(float(data["Epe"]), "kcal/mol").magnitude
            }
            self.calculate_enthalpy_of_formation(tmp_data)
            if "DfE0" in tmp_data:
                value = Q_(tmp_data["DfE0"], "kcal/mol").m_as("eV")
                line += f" REF_energy={value:.4f}"
            else:
                # The energy is the potential, not total energy which includes kinetic
                efact = from_lammps_units(1, "eV").magnitude
                line += f" REF_energy={efact*float(data['Epe']):.4f}"
        if "Sxx" in data:
            sfact = from_lammps_units(1, "eV/Å^3").magnitude
            Sxx = f"{sfact*float(data['Sxx']):.7f}"
            Syy = f"{sfact*float(data['Syy']):.7f}"
            Szz = f"{sfact*float(data['Szz']):.7f}"
            Syz = f"{sfact*float(data['Syz']):.7f}"
            Sxz = f"{sfact*float(data['Sxz']):.7f}"
            Sxy = f"{sfact*float(data['Sxy']):.7f}"
            line += f' REF_stress="{Sxx} {Syy} {Szz} {Syz} {Sxz} {Sxy}" pbc="T T T"'
        else:
            line += ' pbc="F F F"'

        # The time
        if timestep is not None and "timestep" in data:
            time = timestep * data["timestep"]
            line += f' time="{time:.1f~P}"'

        # Add the model, jobserver, and job id if available
        line += f' model="{self.model}"'
        if "SEAMM_JOBSERVER" in os.environ:
            tmp = os.environ["SEAMM_JOBSERVER"]
            line += f' jobserver="{tmp}"'
        if "SEAMM_JOB_ID" in os.environ:
            tmp = os.environ["SEAMM_JOB_ID"]
            line += f' job_id="{tmp}"'

        lines.append(line)

        # And the coordinate lines
        for i, symbol in enumerate(data["symbols"]):
            line = f"{symbol:<2}"
            for val in data["xyz"][i]:
                line += f" {val:14.8f}"
            if "gradients" in data:
                for val in data["gradients"][i]:
                    line += f" {ffact*val:14.8f}"
            if "velocities" in data:
                for val in data["velocities"][i]:
                    line += f" {vfact*val:14.8f}"
            lines.append(line)

        # Add empty line so text ends with newline
        lines.append("")

        return "\n".join(lines)



# -*- coding: utf-8 -*-

"""The graphical part of a LAMMPS step"""

import seamm
import lammps_step


class TkLAMMPS(seamm.TkNode):
    """The node_class is the class of the 'real' node that this
    class is the Tk graphics partner for
    """

    node_class = lammps_step.LAMMPS

    def __init__(
        self,
        tk_flowchart=None,
        node=None,
        namespace="org.molssi.seamm.lammps.tk",
        canvas=None,
        x=None,
        y=None,
        w=200,
        h=50,
    ):
        """Initialize a node

        Keyword arguments:
        """
        self.namespace = namespace

        super().__init__(
            tk_flowchart=tk_flowchart, node=node, canvas=canvas, x=x, y=y, w=w, h=h
        )

        self.create_dialog()

    def create_dialog(self):
        """Create the dialog!"""
        frame = super().create_dialog("Edit LAMMPS Step")

        # make it large!
        sw = self.dialog.winfo_screenwidth()
        sh = self.dialog.winfo_screenheight()
        w = int(0.9 * sw)
        h = int(0.8 * sh)
        x = int(0.05 * sw / 2)
        y = int(0.1 * sh / 2)

        self.dialog.geometry("{}x{}+{}+{}".format(w, h, x, y))

        self.tk_subflowchart = seamm.TkFlowchart(
            master=frame, flowchart=self.node.subflowchart, namespace=self.namespace
        )
        self.tk_subflowchart.draw()

    def right_click(self, event):
        """Probably need to add our dialog..."""

        super().right_click(event)
        self.popup_menu.add_command(label="Edit..", command=self.edit)

        self.popup_menu.tk_popup(event.x_root, event.y_root, 0)

    def update_flowchart(self, tk_flowchart=None, flowchart=None):
        """Update the nongraphical flowchart. Only used in nodes that contain
        flowcharts"""

        super().update_flowchart(
            flowchart=self.node.subflowchart, tk_flowchart=self.tk_subflowchart
        )

    def from_flowchart(self, tk_flowchart=None, flowchart=None):
        """Recreate the graphics from the non-graphical flowchart.
        Only used in nodes that contain flowchart"""

        super().from_flowchart(
            flowchart=self.node.subflowchart, tk_flowchart=self.tk_subflowchart
        )
